# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for sdp_verify.py."""

import random
import unittest

from absl.testing import absltest
from absl.testing import parameterized
from cvxpy.reductions.solvers.defines import INSTALLED_MI_SOLVERS as MIP_SOLVERS
import jax
import jax.numpy as jnp
import jax.scipy
from jax_verify.src.sdp_verify import cvxpy_verify
from jax_verify.src.sdp_verify import eigenvector_utils
from jax_verify.src.sdp_verify import sdp_verify
from jax_verify.src.sdp_verify import utils
from jax_verify.tests.sdp_verify import test_utils
import numpy as np
import optax

NO_MIP_SOLVERS_MESSAGE = 'No mixed-integer solver is installed.'


class LanczosTest(parameterized.TestCase):

  def _test_max_eigenvector_lanczos_once(self, seed=0, dynamic_unroll=False):
    """Test max_eigenvector_lanczos against Scipy eigenvector method."""
    dim = 5
    key = jax.random.PRNGKey(seed)
    h_tmp = jax.random.normal(key, shape=(dim, dim))
    h = h_tmp + jnp.transpose(h_tmp)
    hv = lambda v: jnp.matmul(h, v)
    # Do `dim` iterations of Lanczos
    max_eigenvec_lanczos = eigenvector_utils.max_eigenvector_lanczos(
        hv, dim, dim, key, dynamic_unroll=dynamic_unroll)
    _, eigen_vecs_scipy = jax.scipy.linalg.eigh(h)
    max_eigenvec_scipy = eigen_vecs_scipy[:, -1]
    # Eigenvector can be v or -v
    err = min(jnp.linalg.norm(max_eigenvec_lanczos - max_eigenvec_scipy),
              jnp.linalg.norm(max_eigenvec_lanczos + max_eigenvec_scipy))
    # Eigenvectors have unit norm - check for relative error below 1e-5
    assert err < 1e-5, (f'err: {err}, lanczos: {max_eigenvec_lanczos} '
                        f'scipy: {max_eigenvec_scipy}')

  def test_max_eigenvector_lanczos(self):
    for i in range(10):
      self._test_max_eigenvector_lanczos_once(seed=i)
    for i in range(10):
      self._test_max_eigenvector_lanczos_once(seed=i, dynamic_unroll=True)

  def _test_lanczos_dynamic_vs_static_once(self, seed=0):
    def _safe_div(x1, x2):
      return jnp.where(jnp.logical_and(x1 == 0, x2 == 0), x1, x1 / x2)
    dim = 5
    key = jax.random.PRNGKey(seed)
    h_tmp = jax.random.normal(key, shape=(dim, dim))
    h = h_tmp + jnp.transpose(h_tmp)
    hv = lambda v: jnp.matmul(h, v)
    tr1, vecs1 = eigenvector_utils.lanczos_alg(
        hv, dim, dim, key, dynamic_unroll=True)
    tr2, vecs2 = eigenvector_utils.lanczos_alg(
        hv, dim, dim, key, dynamic_unroll=False)
    assert jnp.max(jnp.abs(_safe_div(tr1 - tr2, tr2))) < 1e-4, (
        f'Seed {seed}: large relative error in Lanczos tridiag')
    assert jnp.max(jnp.abs(_safe_div(vecs1 - vecs2, vecs2))) < 1e-4, (
        f'Seed {seed}: large relative error in Lanczos vecs')

  def test_lanczos_dynamic_vs_static(self):
    for i in range(10):
      self._test_lanczos_dynamic_vs_static_once(seed=i)


class SdpDualPrimalTest(parameterized.TestCase):
  """Tests comparing SDP dual bounds to CVXPY exact solution of primal."""

  @unittest.skipUnless(MIP_SOLVERS, NO_MIP_SOLVERS_MESSAGE)
  def test_crossing_bounds(self):
    loss_margin = 1e-3
    seed = random.randint(1, 10000)
    verif_instance = test_utils.make_toy_verif_instance(seed)
    key = jax.random.PRNGKey(0)
    primal_opt, _ = cvxpy_verify.solve_mip_mlp_elided(verif_instance)
    dual_ub, _ = sdp_verify.solve_sdp_dual(
        utils.make_sdp_verif_instance(verif_instance), key, num_steps=1000)
    assert dual_ub > primal_opt - loss_margin, (
        'Dual upper bound should be greater than optimal primal objective.'
        f'Seed is {seed}. Vals are Dual: {dual_ub} Primal: {primal_opt}')

  def _test_tight_duality_gap(self, seed, loss_margin=0.003, num_steps=3000):
    verif_instance = test_utils.make_toy_verif_instance(
        seed, label=1, target_label=2)
    key = jax.random.PRNGKey(0)
    primal_opt, _ = cvxpy_verify.solve_sdp_mlp_elided(verif_instance)
    dual_ub, _ = sdp_verify.solve_sdp_dual(
        utils.make_sdp_verif_instance(verif_instance), key, num_steps=num_steps,
        verbose=False)
    assert dual_ub - primal_opt < loss_margin, (
        'Primal and dual vals should be close. '
        f'Seed: {seed}. Primal: {primal_opt}, Dual: {dual_ub}')
    assert dual_ub > primal_opt - 1e-3, 'crossing bounds'

  def test_tight_duality_gap(self):
    self._test_tight_duality_gap(0)

  def local_test_tight_duality_gap(self):  # pylint: disable=g-unreachable-test-method
    """Local test, meant to be run in parallel with --tests_per_run."""
    seed = random.randint(1, 10000)
    # 5/300 failures at loss_margin=0.01, 0/300 failures at loss_margin=0.3
    self._test_tight_duality_gap(seed, loss_margin=0.03, num_steps=3000)


class SdpVerifyTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('MLP', 'mlp'),
      ('CNN', 'cnn')
  )
  def test_sdp_dual_simple_no_crash(self, model_type):
    verif_instance = test_utils.make_toy_verif_instance(
        seed=0, target_label=1, label=2, nn=model_type)
    kwargs = {
        'key': jax.random.PRNGKey(0),
        'opt': optax.adam(1e-3),
        'num_steps': 10,
        'eval_every': 5,
        'verbose': False,
        'use_exact_eig_eval': False,
        'use_exact_eig_train': False,
        'n_iter_lanczos': 5,
        'kappa_reg_weight': 1e-5,
        'kappa_zero_after': 8,
        'device_type': None,
    }
    verif_instance = utils.make_sdp_verif_instance(verif_instance)
    # Check all kwargs work.
    dual_val, _ = sdp_verify.solve_sdp_dual_simple(verif_instance, **kwargs)
    assert isinstance(dual_val, float)
    # Check code runs without kwargs.
    dual_val, _ = sdp_verify.solve_sdp_dual_simple(verif_instance, num_steps=5)
    assert isinstance(dual_val, float)

  def test_dual_sdp_no_crash(self):
    for nn in ['cnn', 'mlp']:
      verif_instance = test_utils.make_toy_verif_instance(
          seed=0, target_label=1, label=2, nn=nn)
      key = jax.random.PRNGKey(0)
      dual_val, _ = sdp_verify.solve_sdp_dual(
          utils.make_sdp_verif_instance(verif_instance), key, num_steps=10,
          n_iter_lanczos=5)
    assert isinstance(dual_val, float)

  def test_correct_dual_var_types(self):
    for nn in ['cnn', 'mlp']:
      verif_instance = test_utils.make_toy_verif_instance(
          seed=0, target_label=1, label=2, nn=nn)
      key = jax.random.PRNGKey(0)
      dual_vars = sdp_verify.init_duals(
          utils.make_sdp_verif_instance(verif_instance), key)
      assert len(dual_vars) == 3, 'Input, one hidden layer, kappa'
      assert isinstance(dual_vars[0], sdp_verify.DualVar)
      assert isinstance(dual_vars[1], sdp_verify.DualVarFin)
      assert isinstance(dual_vars[2], jax.interpreters.xla.DeviceArray)

  def test_ibp_init_matches_ibp_bound(self):
    for nn in ['cnn', 'mlp']:
      for seed in range(20):
        orig_verif_instance = test_utils.make_toy_verif_instance(seed, nn=nn)
        key = jax.random.PRNGKey(0)
        verif_instance = utils.make_sdp_verif_instance(orig_verif_instance)
        dual_vars = jax.tree_map(lambda s: None if s is None else jnp.zeros(s),
                                 verif_instance.dual_shapes)
        dual_vars = sdp_verify.init_duals_ibp(verif_instance, dual_vars)
        dual_loss = sdp_verify.dual_fun(
            verif_instance, dual_vars, key, exact=True)
        ibp_bound = utils.ibp_bound_elided(orig_verif_instance)
        assert abs(dual_loss - ibp_bound) < 1e-4, (
            f'Loss at initialization should match IBP: {dual_loss} {ibp_bound}')


class SdpVerifyTestCNNvsMLP(parameterized.TestCase):

  def test_cnn_mlp_match_fixed_window(self):
    num_steps = 1000
    for seed in range(1):
      verif_instance = test_utils.make_toy_verif_instance(seed, nn='cnn_simple')
      key = jax.random.PRNGKey(0)
      params_cnn = verif_instance.params_full
      in_shape = int(np.prod(np.array(params_cnn[0]['W'].shape[:-1])))
      # Input and filter size match -> filter is applied at just one location.
      # Number of layer 1 neurons = no. channels out of conv filter (last dim).
      out_shape = params_cnn[0]['W'].shape[-1]
      params_mlp = [(jnp.reshape(params_cnn[0]['W'],
                                 (in_shape, out_shape)), params_cnn[0]['b']),
                    (params_cnn[1][0], params_cnn[1][1])]
      bounds_mlp = sdp_verify.boundprop(
          params_mlp,
          sdp_verify.IntBound(
              lb=np.zeros((1, in_shape)),
              ub=1 * np.ones((1, in_shape)),
              lb_pre=None,
              ub_pre=None))
      verif_instance_mlp = utils.make_nn_verif_instance(params_mlp, bounds_mlp)
      dual_ub_cnn, _ = sdp_verify.solve_sdp_dual(
          utils.make_sdp_verif_instance(verif_instance), key,
          num_steps=num_steps, verbose=False, use_exact_eig_train=True)
      dual_ub_mlp, _ = sdp_verify.solve_sdp_dual(
          utils.make_sdp_verif_instance(verif_instance_mlp), key,
          num_steps=num_steps, verbose=False, use_exact_eig_train=True)
      assert abs(dual_ub_cnn - dual_ub_mlp) < 1e-2, (
          'Dual upper bound for MLP and CNN (simple CNN) should match.'
          f'Seed is {seed}. Vals are CNN: {dual_ub_cnn} MLP: {dual_ub_mlp}')
      # Error below 1e-4 when run-locally with steps > 3000.
      # Setting error here to 1e-2 with 500 steps for faster unit-tests

  def test_cnn_mlp_match_sliding_window(self):
    num_steps = 1000
    for seed in range(1):
      verif_instance = test_utils.make_toy_verif_instance(seed, nn='cnn_slide')
      key = jax.random.PRNGKey(0)
      lb, ub, params_mlp = test_utils.make_mlp_verif_instance_from_cnn(
          verif_instance)
      bounds_mlp = sdp_verify.boundprop(
          params_mlp,
          sdp_verify.IntBound(lb=lb, ub=ub, lb_pre=None, ub_pre=None))
      verif_instance_mlp = utils.make_nn_verif_instance(params_mlp, bounds_mlp)
      dual_ub_cnn, _ = sdp_verify.solve_sdp_dual(
          utils.make_sdp_verif_instance(verif_instance),
          key,
          num_steps=num_steps,
          verbose=False,
          use_exact_eig_train=True)
      dual_ub_mlp, _ = sdp_verify.solve_sdp_dual(
          utils.make_sdp_verif_instance(verif_instance_mlp),
          key,
          num_steps=num_steps,
          verbose=False,
          use_exact_eig_train=True)
      assert abs(dual_ub_cnn - dual_ub_mlp) < 5e-3, (
          'Dual upper bound for MLP and CNN (sliding filter) should match.'
          f'Seed is {seed}. Vals are CNN: {dual_ub_cnn} MLP: {dual_ub_mlp}')

if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for propagating bounds through the network."""
import functools

from absl.testing import absltest
from absl.testing import parameterized

import haiku as hk
import jax
import jax.numpy as jnp
import jax_verify
from jax_verify.src import bound_propagation
import tree


@hk.without_apply_rng
@hk.transform
def sequential_model(inp):
  net = hk.Sequential([
      hk.Linear(5), jax.nn.relu,
      hk.Linear(2), jax.nn.relu,
      hk.Linear(1)
  ])
  return net(inp)


def residual_model_all_act(inp):
  mod1 = hk.Linear(5)
  mod2 = hk.Linear(5)
  mod3 = hk.Linear(1)

  act_0 = inp
  act_1 = mod1(act_0)
  act_2 = jax.nn.relu(act_1)
  act_3 = mod2(act_2)
  act_4 = jax.nn.relu(act_3)
  act_5 = act_2 + act_4  # Residual
  final_act = mod3(act_5)
  return [act_0, act_1, act_2, act_3, act_4, act_5, final_act]


@hk.without_apply_rng
@hk.transform
def single_element_list_model(inp):
  all_acts = residual_model_all_act(inp)
  return [all_acts[-1]]


@hk.without_apply_rng
@hk.transform
def dict_output_model(inp):
  all_acts = residual_model_all_act(inp)
  return {i: act for i, act in enumerate(all_acts)}


@hk.without_apply_rng
@hk.transform
def residual_model_intermediate(inp):
  return residual_model_all_act(inp)


@hk.without_apply_rng
@hk.transform
def residual_model(inp):
  all_acts = residual_model_all_act(inp)
  return all_acts[-1]


class BoundPropagationTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sequential', sequential_model),
      ('Residual', residual_model)
  )
  def test_model_structure_nostate(self, model):
    z = jnp.array([[1., 2., 3.]])

    params = model.init(jax.random.PRNGKey(1), z)
    input_bounds = jax_verify.IntervalBound(z - 1.0, z + 1.0)
    fun_to_prop = functools.partial(model.apply, params)
    output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, input_bounds)
    self.assertTrue(all(output_bounds.upper >= output_bounds.lower))

  def test_multioutput_model(self):
    z = jnp.array([[1., 2., 3.]])

    fun = hk.without_apply_rng(
        hk.transform(residual_model_all_act, apply_rng=True))
    params = fun.init(jax.random.PRNGKey(1), z)
    input_bounds = jax_verify.IntervalBound(z - 1.0, z + 1.0)
    fun_to_prop = functools.partial(fun.apply, params)
    output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, input_bounds)

    self.assertLen(output_bounds, 7)

  @parameterized.named_parameters(
      ('Sequential', sequential_model),
      ('Residual', residual_model)
  )
  def test_tight_bounds_nostate(self, model):
    z = jnp.array([[1., 2., 3.]])

    params = model.init(jax.random.PRNGKey(1), z)

    tight_input_bounds = jax_verify.IntervalBound(z, z)
    fun_to_prop = functools.partial(model.apply, params)
    tight_output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, tight_input_bounds)

    model_eval = model.apply(params, z)

    # Because the input lower bound is equal to the input upper bound, the value
    # of the output bounds should be the same and correspond to the value of the
    # forward pass.
    self.assertAlmostEqual(tight_output_bounds.lower.tolist(),
                           tight_output_bounds.upper.tolist())

    self.assertAlmostEqual(tight_output_bounds.lower.tolist(),
                           model_eval.tolist())

  @parameterized.named_parameters(
      ('Sequential', sequential_model),
      ('Residual', residual_model),
      ('ResidualAll', residual_model_intermediate),
      ('1elt_list', single_element_list_model),
      ('dict_output', dict_output_model)
  )
  def test_matching_output_structure(self, model):
    def _check_matching_structures(output_tree, bound_tree):
      """Replace all bounds/arrays with True, then compare pytrees."""
      output_struct = tree.traverse(
          lambda x: True if isinstance(x, jnp.ndarray) else None, output_tree)
      bound_struct = tree.traverse(
          lambda x: True if isinstance(x, bound_propagation.Bound) else None,
          bound_tree)
      tree.assert_same_structure(output_struct, bound_struct)
    z = jnp.array([[1., 2., 3.]])
    params = model.init(jax.random.PRNGKey(1), z)
    input_bounds = jax_verify.IntervalBound(z - 1.0, z + 1.0)
    model_output = model.apply(params, z)
    fun_to_prop = functools.partial(model.apply, params)
    for boundprop_method in [
        jax_verify.interval_bound_propagation,
        jax_verify.crown_bound_propagation,
        jax_verify.fastlin_bound_propagation,
        jax_verify.ibpfastlin_bound_propagation,
    ]:
      output_bounds = boundprop_method(fun_to_prop, input_bounds)
      _check_matching_structures(model_output, output_bounds)


class StaticArgumentModel(hk.Module):

  def __init__(self):
    super().__init__()
    self.lin_1 = hk.Linear(5)
    self.lin_2 = hk.Linear(10)

  def __call__(self, inputs, use_2=True):
    if use_2:
      return self.lin_2(inputs)
    else:
      return self.lin_1(inputs)


class StaticArgumentsModelTest(parameterized.TestCase):

  def test_staticargument_last(self):

    @hk.without_apply_rng
    @hk.transform
    def forward(inputs, use_2):
      model = StaticArgumentModel()
      return model(inputs, use_2)

    z = jnp.array([[1., 2., 3.]])
    params = forward.init(jax.random.PRNGKey(1), z, True)
    input_bounds = jax_verify.IntervalBound(z-1.0, z+1.0)

    def fun_to_prop(inputs):
      return forward.apply(params, inputs, True)

    output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, input_bounds)
    self.assertTrue((output_bounds.upper >= output_bounds.lower).all())

  def test_staticargument_first(self):

    @hk.without_apply_rng
    @hk.transform
    def forward(use_2, inputs):
      model = StaticArgumentModel()
      return model(inputs, use_2)

    z = jnp.array([[1., 2., 3.]])
    params = forward.init(jax.random.PRNGKey(1), True, z)
    input_bounds = jax_verify.IntervalBound(z-1.0, z+1.0)

    fun_to_prop = functools.partial(forward.apply, params, True)

    output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, input_bounds)
    self.assertTrue((output_bounds.upper >= output_bounds.lower).all())

  def test_keywords_argument(self):

    @hk.without_apply_rng
    @hk.transform
    def forward(inputs, use_2=False):
      model = StaticArgumentModel()
      return model(inputs, use_2)

    z = jnp.array([[1., 2., 3.]])
    params = forward.init(jax.random.PRNGKey(1), z, use_2=True)
    input_bounds = jax_verify.IntervalBound(z-1.0, z+1.0)
    fun_to_prop = functools.partial(forward.apply, params, use_2=True)
    output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, input_bounds)
    self.assertTrue((output_bounds.upper >= output_bounds.lower).all())


class ModelWithState(hk.Module):

  def __init__(self):
    super().__init__()
    self.lin_1 = hk.Linear(5)
    self.bn_1 = hk.BatchNorm(True, True, decay_rate=0.999)

  def __call__(self, inputs, is_training, test_local_stats=False):
    act = self.lin_1(inputs)
    bn_act = self.bn_1(act, is_training, test_local_stats)
    return bn_act


class StatefulModelTest(parameterized.TestCase):

  def test_stateful_model(self):

    @hk.transform_with_state
    def forward(inputs, is_training, test_local_stats=False):
      model = ModelWithState()
      return model(inputs, is_training, test_local_stats)

    z = jnp.array([[1., 2., 3.]])
    params, state = forward.init(jax.random.PRNGKey(1), z, True, False)

    def fun_to_prop(inputs):
      outs = forward.apply(params, state, jax.random.PRNGKey(1),
                           inputs, False, False)
      # Ignore the outputs that are not the network outputs.
      return outs[0]

    input_bounds = jax_verify.IntervalBound(z-1.0, z+1.0)
    # Consider as static the state, the random generator, and the flags
    output_bounds = jax_verify.interval_bound_propagation(
        fun_to_prop, input_bounds)
    self.assertTrue((output_bounds.upper >= output_bounds.lower).all())


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
# pylint: disable=invalid-name
"""Crown bound propagation used in SDP verification."""

import functools
import jax.numpy as jnp
import jax_verify
from jax_verify.src.sdp_verify import utils
IntBound = utils.IntBound


def boundprop(params, x, epsilon, input_bounds):
  """Runs CROWN-IBP for each layer separately.

  Args:
    params: Parameters for the NN.
    x: Batch of inputs to NN (dimension 2 for MLP or 4 for CNN)
    epsilon: l-inf perturbation to the input.
    input_bounds: Valid lower and upper for the NN as a tuple -- e.g. (0., 1.)

  Returns:
    layer_bounds: upper and lower bounds across the layers of the NN as a list
    of IntBound-s.
  """

  def get_layer_act(layer_idx, inputs):
    act = utils.predict_cnn(params[:layer_idx], inputs)
    return act

  initial_bound = jax_verify.IntervalBound(
      jnp.maximum(x - epsilon, input_bounds[0]),
      jnp.minimum(x + epsilon, input_bounds[1]))

  out_bounds = [IntBound(
      lb_pre=None, ub_pre=None, lb=initial_bound.lower, ub=initial_bound.upper)]
  for i in range(1, len(params) + 1):
    fwd = functools.partial(get_layer_act, i)
    bound = jax_verify.crownibp_bound_propagation(fwd, initial_bound)
    out_bounds.append(
        IntBound(lb_pre=bound.lower, ub_pre=bound.upper,
                 lb=jnp.maximum(0, bound.lower),
                 ub=jnp.maximum(0, bound.upper)))
  return out_bounds



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of Fastlin.
"""
import functools
from typing import Union, Tuple, Optional

import jax
from jax import lax
import jax.numpy as jnp

from jax_verify.src import bound_propagation
from jax_verify.src import ibp
from jax_verify.src import intersection
from jax_verify.src import utils
import numpy as np

Tensor = jnp.ndarray


class LinearExpression:
  """Describes a set of linear expressions."""

  def __init__(self, lin_coeffs, offset):
    """Creates a LinearExpression object.

    Args:
      lin_coeffs: batch x (nb_coeffs) x ... array
      offset: batch x ... array
    """
    self.lin_coeffs = lin_coeffs
    self.offset = offset

  def coeff_dims(self):
    return self.lin_coeffs.shape[:2]

  @property
  def shape(self):
    return self.offset.shape

  def __add__(self, other):
    if isinstance(other, LinearExpression):
      return LinearExpression(self.lin_coeffs + other.lin_coeffs,
                              self.offset + other.offset)
    else:
      return LinearExpression(self.lin_coeffs, self.offset + other)

  def __radd__(self, other):
    return self.__add__(other)

  def __sub__(self, other):
    if isinstance(other, LinearExpression):
      return LinearExpression(self.lin_coeffs - other.lin_coeffs,
                              self.offset - other.offset)
    else:
      return LinearExpression(self.lin_coeffs, self.offset - other)

  def __rsub__(self, other):
    if isinstance(other, LinearExpression):
      return LinearExpression(other.lin_coeffs - self.lin_coeffs,
                              other.offset - self.offset)
    else:
      return LinearExpression(-self.lin_coeffs, other - self.offset,)

  def __truediv__(self, other):
    return LinearExpression(self.lin_coeffs / other, self.offset / other)

  def pack(self) -> Tensor:
    """Return an array suitable for forwarding through a layer.

    This is required because we often have to forward all the lin_coeffs (even
    though they have an extra dimension), and the offset. This function packs
    all elements together in dimension 1 so that we only have to go through the
    module once.

    Use the inverse function `unpack` to recover a LinearExpression

    Returns:
      packed: A tensor with of the shape of the activation at the layer,
        with an additional dimension where all the linear coefficients and the
        offset are concatenated.
    """
    packed = jnp.concatenate([self.lin_coeffs, jnp.expand_dims(self.offset, 1)],
                             axis=1)
    return packed

  @staticmethod
  def unpack(packed_array: Tensor,
             reference: 'LinearExpression') -> 'LinearExpression':
    """Construct a linear expression based on a packed array.

    Packed array might have been obtained from the `pack` function or through
    forwarding a packed array through a layer.

    Args:
      packed_array: Tensor containing a packed array
      reference: Linear expression to copy the coeff dimensions from
    Returns:
      lin_expression: Linear expression
    """
    _, nb_coeffs = reference.coeff_dims()
    lin_coeffs, offset = jnp.split(packed_array, [nb_coeffs], axis=1)
    offset = offset.squeeze(1)
    return LinearExpression(lin_coeffs, offset)


class LinearBound(bound_propagation.Bound):
  """Represent a pair of linear functions that encompass feasible activations.

  We store the linear functions as LinearExpressions objects in `lower_lin` and
  `upper_lin`, and also maintain a reference to the initial bounds on the input
  to be able to concretize the bounds when needed.
  """

  def __init__(self, lower_bound: LinearExpression,
               upper_bound: LinearExpression,
               reference: Optional['LinearBound']):
    self.lower_lin = lower_bound
    self.upper_lin = upper_bound
    self.reference = reference or self
    self._concretized = None

  @property
  def shape(self):
    return self.lower_lin.shape

  @property
  def lower(self):
    if self._concretized is None:
      self.concretize()
    return self._concretized.lower

  @property
  def upper(self):
    if self._concretized is None:
      self.concretize()
    return self._concretized.upper

  def set_concretized(self, interval_bound):
    self._concretized = interval_bound

  @staticmethod
  def initial_linear_bound(lower_bound, upper_bound):
    batch_size = lower_bound.shape[0]
    act_shape = lower_bound.shape[1:]
    input_dim = np.prod(lower_bound.shape[1:])

    sp_lin = jnp.reshape(jnp.eye(input_dim), (input_dim, *act_shape))
    batch_lin = jnp.repeat(jnp.expand_dims(sp_lin, 0), batch_size, axis=0)

    identity_lin = LinearExpression(batch_lin,
                                    jnp.zeros((batch_size, *act_shape)))
    lin_bound = LinearBound(identity_lin, identity_lin, None)
    lin_bound.set_concretized(ibp.IntervalBound(lower_bound, upper_bound))
    return lin_bound

  def concretize(self):
    if self._concretized is not None:
      return self._concretized

    batch_size = self.shape[0]
    nb_act = len(self.shape) - 1
    broad_shape = (batch_size, -1) + (1,)*nb_act
    flat_ref_lb = jnp.reshape(self.reference.lower, broad_shape)
    flat_ref_ub = jnp.reshape(self.reference.upper, broad_shape)

    concrete_lb = (
        (jnp.maximum(self.lower_lin.lin_coeffs, 0.) * flat_ref_lb).sum(axis=1) +
        (jnp.minimum(self.lower_lin.lin_coeffs, 0.) * flat_ref_ub).sum(axis=1) +
        self.lower_lin.offset
    )
    concrete_ub = (
        (jnp.maximum(self.upper_lin.lin_coeffs, 0.) * flat_ref_ub).sum(axis=1) +
        (jnp.minimum(self.upper_lin.lin_coeffs, 0.) * flat_ref_lb).sum(axis=1) +
        self.upper_lin.offset
    )

    self._concretized = ibp.IntervalBound(concrete_lb, concrete_ub)
    return self._concretized

  def __add__(self, other):
    if isinstance(other, LinearBound):
      if self.reference is not other.reference:
        raise ValueError('Adding bounds referring to different inputs.')
      return LinearBound(self.lower_lin + other.lower_lin,
                         self.upper_lin + other.upper_lin,
                         self.reference)
    else:
      return LinearBound(self.lower_lin + other,
                         self.upper_lin + other,
                         self.reference)

  def __radd__(self, other):
    return self.__add__(other)

  def __sub__(self, other):
    if isinstance(other, LinearBound):
      if self.reference is not other.reference:
        raise ValueError('Adding bounds refering to different inputs.')
      return LinearBound(self.lower_lin - other.upper_lin,
                         self.upper_lin - other.lower_lin,
                         self.reference)
    else:
      return LinearBound(self.lower_lin - other,
                         self.upper_lin - other,
                         self.reference)

  def __rsub__(self, other):
    if isinstance(other, LinearBound):
      if self.reference is not other.reference:
        raise ValueError('Adding bounds refering to different inputs.')
      return LinearBound(other.lower_lin - self.upper_lin,
                         other.upper_lin - self.lower_lin,
                         self.reference)
    else:
      return LinearBound(other - self.upper_lin,
                         other - self.lower_lin,
                         self.reference)


def _decompose_affine_argument(hs: Union[LinearBound, Tensor]
                               ) -> Tuple[Tensor, Tensor,
                                          Optional[LinearBound]]:
  """Decompose an argument for a (bound, parameter) Fastlin propagation.

  We do not need to know which argument is the bound and which one is the
  parameter because we can simply decompose the operation.

  To propagate W * x, we can simply do
    out_diff = abs(W) * (x.upper - x.lower) / 2
    out_mean = W * (x.upper + x.lower) / 2
    out_ub = out_mean + out_diff
    out_lb = out_mean - out_diff

  This function will separate W and x in the two components we need so that we
  do not have to worry about argument order.

  Args:
    hs: Either a LinearBound or a tensor, input argument to an affine function.
  Returns:
    part1: Packed coefficients for the bound "range" if hs is a bound, or
      absolute value of the tensor if hs is a tensor.
    part2: Packed coefficients for the bound "mean" if hs is a bound, or
      value of the tensor if hs is a tensor.
    part3: Reference for the bound if hs is a bound, otherwise None.
  """
  if isinstance(hs, LinearBound):
    return (((hs.upper_lin - hs.lower_lin) / 2).pack(),
            ((hs.upper_lin + hs.lower_lin) / 2).pack(),
            hs.reference)
  else:
    return jnp.abs(hs), hs, None


@bound_propagation.unwrapping
def _fastlin_linear_op(lin_fun, lhs, rhs, **kwargs):
  """Propagation of Linear bounds through the linear operation `lin_fun`.

  Args:
    lin_fun: Linear function to pass through.
    lhs: Either parameters or LinearBound.
    rhs: Either parameters or LinearBound.
    **kwargs: Dict with the parameters of the linear operation.
  Returns:
    out_bounds: LinearBound
  """
  if isinstance(lhs, LinearBound) != isinstance(rhs, LinearBound):

    lhs_0, lhs_1, ref_l = _decompose_affine_argument(lhs)
    rhs_0, rhs_1, ref_r = _decompose_affine_argument(rhs)
    # One of the reference is None, so we keep only the one defined.
    ref = ref_l or ref_r

    vmap_axis = tuple(1 if isinstance(arg, LinearBound) else None
                      for arg in [lhs, rhs])
    unkwarged_linfun = functools.partial(lin_fun, **kwargs)
    vmap_linfun = jax.vmap(unkwarged_linfun, in_axes=vmap_axis, out_axes=1)

    unp_forward_mean = vmap_linfun(lhs_1, rhs_1)
    unp_forward_range = vmap_linfun(lhs_0, rhs_0)

    # pytype: disable=attribute-error
    # ref is necessarily a Bound, because either lhs or rhs is a bound, so
    # it's not possible for `ref_l` and `ref_r` to both be None, but pytype
    # can not deduce that and will raise an attribute-error for lower_lin.
    forward_mean = LinearExpression.unpack(unp_forward_mean, ref.lower_lin)
    forward_range = LinearExpression.unpack(unp_forward_range, ref.lower_lin)
    # pytype: enable=attribute-error
    out_lb = forward_mean - forward_range
    out_ub = forward_mean + forward_range

    return LinearBound(out_lb, out_ub, ref)
  elif ((not isinstance(lhs, bound_propagation.Bound)) and
        (not isinstance(rhs, bound_propagation.Bound))):
    # Both are arrays, so can simply go through
    return lin_fun(lhs, rhs, **kwargs)
  else:
    raise ValueError('BoundPropagation through Linear operation '
                     'is not supported when both inputs are bounds.')


@bound_propagation.unwrapping
def _fastlin_add(lhs, rhs):
  return lhs + rhs


@bound_propagation.unwrapping
def _fastlin_sub(lhs, rhs):
  return lhs - rhs


def _fastlin_div(lhs, rhs):
  """Propagation of Linear bounds through Elementwise division.

  We don't support the propagation of bounds through the denominator.

  Args:
    lhs: Numerator of the division.
    rhs: Denominator of the division.
  Returns:
    out_bounds: Bound on the output of the division.
  """
  if isinstance(rhs, bound_propagation.Bound):
    raise ValueError('Bound propagation through the denominator unsupported.')
  return _fastlin_linear_op(lambda x, y: x/y, lhs, rhs)


def _fastlin_max(lhs, rhs):
  """Propagation of Fastlin bounds through a max.

  At the moment, only the ReLU is supported.

  This relaxes the ReLU with the parallel bounds of slope (ub) / (ub - lb)

  Args:
    lhs: First input to the max, assumed to be a ReLU input
    rhs: Second input to the max, assumed to be 0
  Returns:
    out_bounds: FastlinBound.
  """
  if not (isinstance(lhs, bound_propagation.Bound) and rhs == 0.):
    raise NotImplementedError('Only ReLU is implemented for now.')

  # Incorporate applied concrete bounds from the wrapper, if any.
  lhs_lower, lhs_upper = lhs.lower, lhs.upper
  lhs = lhs.unwrap()

  relu_on = (lhs_lower >= 0.)
  relu_amb = jnp.logical_and(lhs_lower < 0., lhs_upper >= 0.)
  slope = relu_on.astype(jnp.float32)
  slope += jnp.where(relu_amb,
                     lhs_upper / jnp.maximum(lhs_upper - lhs_lower, 1e-12),
                     jnp.zeros_like(lhs_lower))
  ub_offset = jnp.where(relu_amb, - slope * lhs_lower,
                        jnp.zeros_like(lhs_lower))

  broad_slope = jnp.expand_dims(slope, 1)
  ub_lin_coeffs = broad_slope * lhs.upper_lin.lin_coeffs
  lb_lin_coeffs = broad_slope * lhs.lower_lin.lin_coeffs
  ub_lin_offset = slope * lhs.upper_lin.offset + ub_offset
  lb_lin_offset = slope * lhs.lower_lin.offset

  return LinearBound(LinearExpression(lb_lin_coeffs, lb_lin_offset),
                     LinearExpression(ub_lin_coeffs, ub_lin_offset),
                     lhs.reference)


def _crown_max(lhs, rhs):
  """Propagation of Fastlin bounds through a max.

  This relaxes the ReLU with the adaptive choice of lower bounds as described
  for CROWN-ada in https://arxiv.org/abs/1811.00866.

  Args:
    lhs: First input to the max, assumed to be a ReLU input
    rhs: Second input to the max, assumed to be 0
  Returns:
    out_bounds: FastlinBound.
  """
  if not (isinstance(lhs, bound_propagation.Bound) and rhs == 0.):
    raise NotImplementedError('Only ReLU is implemented for now.')

  # Incorporate applied concrete bounds from the wrapper, if any.
  lhs_lower, lhs_upper = lhs.lower, lhs.upper
  lhs = lhs.unwrap()

  relu_on = (lhs_lower >= 0.)
  relu_amb = jnp.logical_and(lhs_lower < 0., lhs_upper >= 0.)
  ub_slope = relu_on.astype(jnp.float32)
  ub_slope += jnp.where(relu_amb,
                        lhs_upper / jnp.maximum(lhs_upper - lhs_lower, 1e-12),
                        jnp.zeros_like(lhs_lower))
  ub_offset = jnp.where(relu_amb, - ub_slope * lhs_lower,
                        jnp.zeros_like(lhs_lower))
  lb_slope = (ub_slope >= 0.5).astype(jnp.float32)

  broad_ub_slope = jnp.expand_dims(ub_slope, 1)
  broad_lb_slope = jnp.expand_dims(lb_slope, 1)
  ub_lin_coeffs = broad_ub_slope * lhs.upper_lin.lin_coeffs
  lb_lin_coeffs = broad_lb_slope * lhs.lower_lin.lin_coeffs
  ub_lin_offset = ub_slope * lhs.upper_lin.offset + ub_offset
  lb_lin_offset = lb_slope * lhs.lower_lin.offset

  return LinearBound(LinearExpression(lb_lin_coeffs, lb_lin_offset),
                     LinearExpression(ub_lin_coeffs, ub_lin_offset),
                     lhs.reference)


@bound_propagation.unwrapping
def _fastlin_coeffwise(
    primitive: jax.core.Primitive,
    *args: Union[LinearBound, jnp.ndarray],
    **params
    ) -> LinearBound:
  """Propagation of Fastlin bounds through a coefficient-wise op like reshape.

  This vectorises the given primitive to apply it to each coefficient of the
  bounds' linear expressions.

  Args:
    primitive: Operation to apply to the Fastlin bounds.
    *args: Arguments of the primitive.
    **params: Keyword arguments of the primitive.
  Returns:
    out_bounds: FastlinBound, reshaped.
  """
  if any(isinstance(arg, LinearBound) for arg in args):
    bound_arg = [arg for arg in args if isinstance(arg, LinearBound)][0]

    vmap_axis = tuple(1 if isinstance(arg, LinearBound) else None
                      for arg in args)
    op = functools.partial(primitive.bind, **params)
    vmap_op = jax.vmap(op, in_axes=vmap_axis, out_axes=1)

    return LinearBound(
        LinearExpression(
            vmap_op(*[arg.lower_lin.lin_coeffs if isinstance(arg, LinearBound)
                      else arg for arg in args]),
            op(*[arg.lower_lin.offset if isinstance(arg, LinearBound)
                 else arg for arg in args])),
        LinearExpression(
            vmap_op(*[arg.upper_lin.lin_coeffs if isinstance(arg, LinearBound)
                      else arg for arg in args]),
            op(*[arg.upper_lin.offset if isinstance(arg, LinearBound)
                 else arg for arg in args])),
        reference=bound_arg.reference)

  else:
    # No interval dependence. Compute the output value directly.
    return primitive.bind(*args, **params)


fastlin_input_transform = (
    lambda _, lower, upper: LinearBound.initial_linear_bound(lower, upper))
fastlin_primitive_transform = {
    lax.add_p: _fastlin_add,
    lax.sub_p: _fastlin_sub,
    lax.max_p: _fastlin_max,
    lax.reshape_p: functools.partial(_fastlin_coeffwise, lax.reshape_p),
    lax.reduce_sum_p: functools.partial(_fastlin_coeffwise, lax.reduce_sum_p),
    lax.dot_general_p: functools.partial(_fastlin_linear_op, lax.dot_general),
    lax.conv_general_dilated_p: functools.partial(
        _fastlin_linear_op, utils.wrapped_general_conv),
    lax.mul_p: functools.partial(_fastlin_linear_op, lambda x, y: x*y),
    lax.div_p: _fastlin_div,
}
fastlin_primitive_transform = {
    key: utils.simple_propagation(fastlin_primitive_transform[key])
    for key in fastlin_primitive_transform
}
fastlin_transform = bound_propagation.OpwiseBoundTransform(
    fastlin_input_transform, fastlin_primitive_transform)


def fastlin_bound_propagation(function, *bounds):
  """Performs FastLin as described in https://arxiv.org/abs/1804.09699.

  Args:
    function: Function performing computation to obtain bounds for. Takes as
      only argument the network inputs.
    *bounds: jax_verify.IntervalBound, bounds on the inputs of the function.
  Returns:
    output_bound: Bounds on the output of the function obtained by FastLin
  """
  output_bound, _ = bound_propagation.bound_propagation(
      fastlin_transform, function, *bounds)
  return output_bound


_crown_input_transform = fastlin_input_transform
_crown_primitive_transform = fastlin_primitive_transform.copy()
_crown_primitive_transform.update({
    lax.max_p: utils.simple_propagation(_crown_max),
})
_crown_transform = bound_propagation.OpwiseBoundTransform(
    _crown_input_transform, _crown_primitive_transform)


def crown_bound_propagation(function, *bounds):
  """Performs CROWN as described in https://arxiv.org/abs/1811.00866.

  Args:
    function: Function performing computation to obtain bounds for. Takes as
      only argument the network inputs.
    *bounds: jax_verify.IntervalBound, bounds on the inputs of the function.
  Returns:
    output_bound: Bounds on the output of the function obtained by FastLin
  """
  output_bound, _ = bound_propagation.bound_propagation(
      _crown_transform, function, *bounds)
  return output_bound


def ibpfastlin_bound_propagation(function, *bounds):
  """Obtains the best of IBP and Fastlin bounds.

  Args:
    function: Function performing computation to obtain bounds for. Takes as
      only argument the network inputs.
    *bounds: jax_verify.IntervalBound, bounds on the inputs of the function.
  Returns:
    output_bound: Bounds on the output of the function obtained by FastLin
  """
  output_bound, _ = bound_propagation.bound_propagation(
      intersection.IntersectionBoundTransform(
          ibp.bound_transform, fastlin_transform),
      function, *bounds)
  return output_bound



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library to perform verificaton on Neural Networks.
"""

from jax_verify.src.crown_ibp import crownibp_bound_propagation
from jax_verify.src.cvxpy_relaxation_solver import CvxpySolver
from jax_verify.src.fastlin import crown_bound_propagation
from jax_verify.src.fastlin import fastlin_bound_propagation
from jax_verify.src.fastlin import fastlin_transform
from jax_verify.src.fastlin import ibpfastlin_bound_propagation
from jax_verify.src.fastlin import LinearBound
from jax_verify.src.ibp import bound_transform as ibp_transform
from jax_verify.src.ibp import interval_bound_propagation
from jax_verify.src.ibp import IntervalBound
from jax_verify.src.intersection import IntersectionBoundTransform
from jax_verify.src.solve_relaxation import solve_planet_relaxation
from jax_verify.src.utils import open_file



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
# pylint: disable=invalid-name
"""Neural network verification with cvxpy for correctness checks."""

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_MI_SOLVERS as MIP_SOLVERS
import jax.numpy as jnp
from jax_verify.src.sdp_verify import utils
import numpy as np
import scipy


def solve_mip_mlp_elided(verif_instance):
  """Compute optimal attack loss for MLPs, via exactly solving MIP."""
  assert MIP_SOLVERS, 'No MIP solvers installed with cvxpy.'
  assert verif_instance.type == utils.VerifInstanceTypes.MLP_ELIDED
  params, bounds, obj, obj_const = (
      verif_instance.params, verif_instance.bounds, verif_instance.obj,
      verif_instance.const)
  layer_sizes = utils.mlp_layer_sizes(params)
  on_state = []
  post_activations = [cp.Variable((1, layer_sizes[0]))]
  pre_activations = []
  constraints = []

  for (i, param) in enumerate(params):
    W, b = param
    b = jnp.reshape(b, (1, b.size))
    on_state.append(cp.Variable((1, b.size), boolean=True))
    pre_activations.append(cp.Variable((1, b.size)))
    post_activations.append(cp.Variable((1, b.size)))

    # Linear relaxation of ReLU constraints
    constraints += [pre_activations[-1] == post_activations[-2]@W + b]
    constraints += [post_activations[-1] >= pre_activations[-1]]
    constraints += [post_activations[-1] >= 0]

    # If ReLU is off, post activation is non-positive. Otherwise <= ub
    constraints += [post_activations[-1] <= cp.multiply(on_state[-1],
                                                        bounds[i+1].ub)]

    # If ReLU is off, pre-activation is non-positive. Otherwise <= ub_pre
    constraints += [pre_activations[-1] <= cp.multiply(on_state[-1],
                                                       bounds[i+1].ub_pre)]

    # If ReLU is on, post-activation == pre-activation
    # Define <= here, >= constraint added above.
    constraints += [post_activations[-1]-pre_activations[-1] <=
                    cp.multiply(1-on_state[-1],
                                bounds[i+1].ub-bounds[i+1].lb_pre)]

  # Optionally, include IBP bounds to speed up MIP solving
  # Post activations are within bounds
  # i=0 case encodes input constraint
  for (i, post) in enumerate(post_activations):
    constraints += [post <= bounds[i].ub]
    constraints += [post >= bounds[i].lb]

  # # Pre activations are within bounds
  for (i, pre) in enumerate(pre_activations):
    constraints += [pre <= bounds[i+1].ub_pre]
    constraints += [pre >= bounds[i+1].lb_pre]

  # Set objective over final post-activations
  obj_cp = cp.sum(cp.multiply(obj, post_activations[-1]))

  # Define and solve problem
  problem = cp.Problem(cp.Maximize(obj_cp), constraints)
  # NB: Originally, we used cp.ECOS_BB here, but cvxpy 1.1 drops support,
  # so we just use the first available MIP solver (which is dependent on user
  # installation).
  problem.solve(solver=MIP_SOLVERS[0])

  # Report results
  info = {
      'problem': problem,
      'post': post_activations,
      'pre': pre_activations,
  }
  return obj_cp.value + obj_const, info


def solve_lp_primal_elided(verif_instance):
  """Compute optimal attack loss for MLPs against LP relaxation."""
  assert verif_instance.type == utils.VerifInstanceTypes.MLP_ELIDED
  params, bounds, obj, obj_const = (
      verif_instance.params, verif_instance.bounds, verif_instance.obj,
      verif_instance.const)
  layer_sizes = utils.mlp_layer_sizes(params)
  post_activations = [cp.Variable((1, layer_sizes[0]))]
  constraints = []

  for (i, param) in enumerate(params):
    W, b = param
    b = jnp.reshape(b, (1, b.size))
    post_activations.append(cp.Variable((1, b.size)))
    pre_act = post_activations[-2]@W + b
    post_act = post_activations[-1]

    # Linear relaxation of ReLU constraints
    constraints += [post_act >= pre_act]
    constraints += [post_act >= 0]

    # Triangle relaxation
    l = np.minimum(0., bounds[i+1].lb_pre)
    u = np.maximum(0., bounds[i+1].ub_pre)
    constraints += [cp.multiply(u, pre_act) - cp.multiply(u, l) -
                    cp.multiply(u - l, post_act) >= 0]

  # Optionally, include IBP bounds to speed up MIP solving
  # Post activations are within bounds
  # i=0 case encodes input constraint
  for (i, post) in enumerate(post_activations[:1]):
    constraints += [post <= bounds[i].ub]
    constraints += [post >= bounds[i].lb]

  # Set objective over final post-activations
  obj_cp = cp.sum(cp.multiply(obj, post_activations[-1]))

  # Define and solve problem
  problem = cp.Problem(cp.Maximize(obj_cp), constraints)
  problem.solve(solver=cp.ECOS)

  # Report results
  info = {
      'problem': problem,
      'post': post_activations,
  }
  return obj_cp.value + obj_const, info


def solve_sdp_mlp_elided(verif_instance, solver_name='SCS', verbose=False,
                         check_feasibility=False, feasibility_margin=0.0):
  """Compute exact SDP relaxation verified bound, following Raghunathan 18.

  Args:
    verif_instance: VerifInstance namedtuple
    solver_name: string, SDP solver, either 'SCS' or 'CVXOPT'
    verbose: bool, controls verbose output from SDP solver
    check_feasibility: bool, if True, try to find any verified certificate,
      rather than tightest possible lower bound
    feasibility_margin: float, when `check_feasibility=True`, verify that
      adversary cannot decrease objective below `feasibility_margin`.

  Returns:
    obj_value: either a float, the bound on objective (check_feasibility=False),
      or a bool, whether verification succeeded (check_feasibility=True)
    info: dict of other info, e.g. solver status, values found by solver
  """
  assert verif_instance.type == utils.VerifInstanceTypes.MLP_ELIDED
  params, input_bounds, bounds, obj, obj_const = (
      verif_instance.params, verif_instance.input_bounds, verif_instance.bounds,
      verif_instance.obj, verif_instance.const)
  layer_sizes = utils.mlp_layer_sizes(params)
  assert len(bounds) == len(layer_sizes) + 1
  # Matrix P, where P = vv' before SDP relaxation, and v = [1, x_0, ..., x_L]
  P = cp.Variable((1 + sum(layer_sizes), 1 + sum(layer_sizes)))

  # Matrix constraints
  constraints = [
      P == P.T,
      P >> 0,
      P[0][0] == 1.,
  ]

  cumsum_sizes = [0] + list(np.cumsum(layer_sizes))
  def _slice(i):
    """Helper method for `p_slice`."""
    if i == -1:
      return 0
    else:
      return slice(1 + cumsum_sizes[i], 1 + cumsum_sizes[i+1])

  def p_slice(i, j):
    """Symbolic indexing into matrix P.

    Args:
      i: an integer, either -1, or in [0, num_layers).
      j: an integer, either -1, or in [0, num_layers).

    Returns:
      slice object, used to index into P. In the QCQP, if P = vv', where
        v = [1, x_0, ..., x_L], then p_slice(i, j) gives submatrix corresponding
        to P[x_i x_j']. When j = -1, this returns P[x_i].
    """
    return P[_slice(i), _slice(j)]
  diag = cp.atoms.affine.diag.diag

  # Input/IBP constraints
  # TODO: Check if these are actually necessary
  if input_bounds is not None:
    constraints += [p_slice(0, -1) >= input_bounds[0]]
    constraints += [p_slice(0, -1) <= input_bounds[1]]

  for i in range(len(layer_sizes)):
    lb = bounds[i].lb[0]
    ub = bounds[i].ub[0]
    assert lb.shape == ub.shape == (layer_sizes[i],)
    assert diag(p_slice(i, i)).shape == (layer_sizes[i],)
    assert p_slice(i, -1).shape == (layer_sizes[i],)
    constraints += [diag(p_slice(i, i)) <=
                    cp.multiply(lb + ub, p_slice(i, -1)) - cp.multiply(lb, ub)]

  # Relu / weight constraints
  for i, param in enumerate(params):
    W, b = param
    constraints += [p_slice(i+1, -1) >= 0]
    constraints += [p_slice(i+1, -1) >= p_slice(i, -1)@W + b]
    # Encode constraint P[zz'] = WP[xz']. Since our networks use z=xW+b rather
    # than z=Wx+b, we use W'x = xW (which holds when x is a vector)
    constraints += [
        diag(p_slice(i+1, i+1)) ==
        diag(W.T@p_slice(i, i+1)) + cp.multiply(b, p_slice(i+1, -1))]

  # Set objective over final post-activations
  final_idx = len(layer_sizes)-1
  x_final = P[_slice(final_idx), _slice(-1)]
  obj_cp = cp.sum(cp.multiply(obj[0], x_final))

  # Define and solve problem
  if check_feasibility:
    constraints += [obj_cp + obj_const >= feasibility_margin]
    problem = cp.Problem(cp.Maximize(cp.Constant(0.)), constraints)
  else:
    problem = cp.Problem(cp.Maximize(obj_cp), constraints)
  solver = getattr(cp, solver_name)
  problem.solve(solver=solver, verbose=verbose)

  # Report results
  info = {
      'problem': problem,
      'P': P,
      'constraints': constraints,
  }
  print('status', problem.status)
  if check_feasibility:
    # If solver shows problem is infeasible for adversary, this is a certificate
    obj_value = problem.status == 'infeasible'
  else:
    obj_value = obj_cp.value + obj_const if obj_cp.value is not None else -99999
  return obj_value, info


def _violation(arr):
  return np.maximum(0, (np.max(arr)))


def _violation_leq(arr1, arr2):
  """Get violation for constraint `arr1 <= arr2`."""
  return _violation(arr1 - arr2)


def check_sdp_bounds_numpy(P, verif_instance, input_bounds=(0, 1)):
  """Check SDP solution for 1-hidden MLP satisfies constraints in numpy."""
  params, bounds, obj, const = (
      verif_instance.params, verif_instance.bounds, verif_instance.obj,
      verif_instance.const)
  layer_sizes = utils.mlp_layer_sizes(params)
  assert len(layer_sizes) == 2, 'Relu MLP with 1 hidden layer'
  assert len(params) == 1, 'Relu MLP with 1 hidden layer'
  assert P.shape == (1+sum(layer_sizes), 1+sum(layer_sizes))
  violations = {}
  # Matrix constraints
  violations['P = P.T'] = _violation(np.abs(P - P.T))
  violations['P[0][0] = 1'] = abs(P[0][0] - 1.0)
  eig_vals, _ = scipy.linalg.eigh(P)
  violations['P >= 0 (SDP)'] = _violation(-eig_vals)

  x = P[0, 1:1+layer_sizes[0]]
  z = P[0, 1+layer_sizes[0]:]
  xx = P[1:1+layer_sizes[0], 1:1+layer_sizes[0]]
  xz = P[1:1+layer_sizes[0], 1+layer_sizes[0]:]
  zz = P[1+layer_sizes[0]:, 1+layer_sizes[0]:]

  # Relu constraints
  w, b = params[0]
  violations['relu_0'] = _violation_leq(0, z)
  violations['relu_wx_b'] = _violation_leq(np.matmul(x, w) + b, z)
  violations['relu_eq'] = _violation(np.abs(
      np.diag(np.matmul(w.T, xz)) + b*z - np.diag(zz)))

  # Input bound constraints
  violations['input_lb'] = _violation_leq(input_bounds[0], x)
  violations['input_ub'] = _violation_leq(x, input_bounds[1])

  # Interval bound constraints
  for i in range(len(layer_sizes)):
    lb = bounds[i].lb[0]
    ub = bounds[i].ub[0]
    x_slice = slice(1+sum(layer_sizes[:i]), 1+sum(layer_sizes[:i+1]))
    x = P[0, x_slice]
    xx = P[x_slice, x_slice]
    violations[f'lay{i}_bound'] = _violation_leq(np.diag(xx), (lb+ub)*x - lb*ub)

  # Objective
  obj = const + np.sum(obj * z)
  return obj, violations



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Interval Bound Propagation."""

import functools

from absl.testing import absltest
from absl.testing import parameterized

import haiku as hk
import jax
import jax.numpy as jnp
import jax_verify


class IBPTest(parameterized.TestCase):

  def assertArrayAlmostEqual(self, lhs, rhs):
    diff = jnp.abs(lhs - rhs).max()
    self.assertAlmostEqual(diff, 0.)

  def test_linear_ibp(self):

    def linear_model(inp):
      return hk.Linear(1)(inp)

    z = jnp.array([[1., 2., 3.]])
    params = {'linear':
              {'w': jnp.ones((3, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(
        hk.without_apply_rng(hk.transform(linear_model, apply_rng=True)).apply,
        params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.interval_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(5., output_bounds.lower)
    self.assertAlmostEqual(11., output_bounds.upper)

  def test_conv1d_ibp(self):

    def conv1d_model(inp):
      return hk.Conv1D(output_channels=1, kernel_shape=2,
                       padding='VALID', stride=1, with_bias=True)(inp)
    z = jnp.array([3., 4.])
    z = jnp.reshape(z, [1, 2, 1])

    params = {'conv1_d':
              {'w': jnp.ones((2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(
        hk.without_apply_rng(hk.transform(conv1d_model, apply_rng=True)).apply,
        params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.interval_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(7., output_bounds.lower)
    self.assertAlmostEqual(11., output_bounds.upper)

  def test_conv2d_ibp(self):
    def conv2d_model(inp):
      return hk.Conv2D(output_channels=1, kernel_shape=(2, 2),
                       padding='VALID', stride=1, with_bias=True)(inp)
    z = jnp.array([1., 2., 3., 4.])
    z = jnp.reshape(z, [1, 2, 2, 1])

    params = {'conv2_d':
              {'w': jnp.ones((2, 2, 1, 1), dtype=jnp.float32),
               'b': jnp.array([2.])}}

    fun = functools.partial(
        hk.without_apply_rng(hk.transform(conv2d_model, apply_rng=True)).apply,
        params)
    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.interval_bound_propagation(fun, input_bounds)

    self.assertAlmostEqual(8., output_bounds.lower)
    self.assertAlmostEqual(16., output_bounds.upper)

  def test_relu_ibp(self):
    def relu_model(inp):
      return jax.nn.relu(inp)
    z = jnp.array([[-2., 3.]])

    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.interval_bound_propagation(relu_model,
                                                          input_bounds)

    self.assertArrayAlmostEqual(jnp.array([[0., 2.]]), output_bounds.lower)
    self.assertArrayAlmostEqual(jnp.array([[0., 4.]]), output_bounds.upper)

  def test_softplus_ibp(self):
    def softplus_model(inp):
      return jax.nn.softplus(inp)
    z = jnp.array([[-2., 3.]])

    input_bounds = jax_verify.IntervalBound(z - 1., z + 1.)
    output_bounds = jax_verify.interval_bound_propagation(softplus_model,
                                                          input_bounds)

    self.assertArrayAlmostEqual(jnp.logaddexp(z - 1., 0),
                                output_bounds.lower)
    self.assertArrayAlmostEqual(jnp.logaddexp(z + 1., 0),
                                output_bounds.upper)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for crown_boundprop.py."""

import pickle
from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
import jax_verify
from jax_verify.src.sdp_verify import crown_boundprop
from jax_verify.src.sdp_verify import utils
import numpy as np


class BoundpropTest(parameterized.TestCase):

  def test_crown_boundprop(self):
    """Test CROWN bounds vs FGSM on Wong-Small MNIST CNN."""
    self._test_boundprop(crown_boundprop.boundprop)

  def test_ibp_boundprop(self):
    def boundprop(params, x, epsilon, input_bounds):
      assert len(x.shape) == 4 and x.shape[0] == 1, f'shape check {x.shape}'
      init_bound = utils.init_bound(x[0], epsilon, input_bounds=input_bounds)
      return utils.boundprop(params, init_bound)
    self._test_boundprop(boundprop)

  def _test_boundprop(self, boundprop_method):
    """Test `boundprop_method` on Wong-Small MNIST CNN."""
    with jax_verify.open_file('mnist/x_test_first100.npy', 'rb') as f:
      xs = np.load(f)
    model_name = 'models/mnist_wongsmall_eps_10_adv.pkl'
    with jax_verify.open_file(model_name, 'rb') as f:
      params = pickle.load(f)
    x = xs[0]
    eps = 0.1

    bounds = boundprop_method(params, np.expand_dims(x, axis=0), eps,
                              input_bounds=(0., 1.))
    crown_lbs = utils.flatten([b.lb_pre for b in bounds[1:]])
    crown_ubs = utils.flatten([b.ub_pre for b in bounds[1:]])

    num_idxs_to_test = 10
    max_idx = crown_lbs.shape[0]
    np.random.seed(0)
    test_idxs = np.random.randint(max_idx, size=num_idxs_to_test)

    @jax.jit
    def fwd(x):
      _, acts = utils.predict_cnn(params, jnp.expand_dims(x, 0),
                                  include_preactivations=True)
      return acts

    get_act = lambda x, idx: utils.flatten(fwd(x), backend=jnp)[idx]

    print('Number of activations:', crown_lbs.shape[0])
    print('Bound shape', [b.lb.shape for b in bounds])
    print('Activation shape', [a.shape for a in fwd(x)])
    assert utils.flatten(fwd(x)).shape == crown_lbs.shape, (
        f'bad shape {crown_lbs.shape}, {utils.flatten(fwd(x)).shape}')

    for idx in test_idxs:
      nom = get_act(x, idx)
      crown_lb = crown_lbs[idx]
      crown_ub = crown_ubs[idx]

      adv_loss = lambda x: get_act(x, idx)  # pylint: disable=cell-var-from-loop
      x_lb = utils.pgd(adv_loss, x, eps, 50, 0.01)
      fgsm_lb = get_act(x_lb, idx)

      adv_loss = lambda x: -get_act(x, idx)  # pylint: disable=cell-var-from-loop
      x_ub = utils.pgd(adv_loss, x, eps, 50, 0.01)
      fgsm_ub = get_act(x_ub, idx)

      print(f'Idx {idx}: Boundprop LB {crown_lb}, FGSM LB {fgsm_lb}, '
            f'Nominal {nom}, FGSM UB {fgsm_ub}, Boundprop UB {crown_ub}')
      margin = 1e-5
      assert crown_lb <= fgsm_lb + margin, f'Bad lower bound. Idx {idx}.'
      assert crown_ub >= fgsm_ub - margin, f'Bad upper bound. Idx {idx}.'

      crown_lb_post, fgsm_lb_post = max(crown_lb, 0), max(fgsm_lb, 0)
      crown_ub_post, fgsm_ub_post = max(crown_ub, 0), max(fgsm_ub, 0)
      assert crown_lb_post <= fgsm_lb_post + margin, f'Idx {idx}.'
      assert crown_ub_post >= fgsm_ub_post - margin, f'Idx {idx}.'


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils used for JAX neural network verification."""

import os
from jax import lax
import urllib.request

######## File Loading ########

def open_file(name, *open_args, **open_kwargs):
  """Load file, downloading to /tmp/jax_verify first if necessary."""
  local_root = '/tmp/jax_verify'
  local_path = os.path.join(local_root, name)
  if not os.path.exists(os.path.dirname(local_path)):
    os.makedirs(os.path.dirname(local_path))
  if not os.path.exists(local_path):
    gcp_bucket_url = 'https://storage.googleapis.com/deepmind-jax-verify/'
    download_url = gcp_bucket_url + name
    urllib.request.urlretrieve(download_url, local_path)
  return open(local_path, *open_args, **open_kwargs)

######### Miscellaneous #########


def collect_required_arguments(req_args, all_kwargs):
  """Extract a dictionary with the required keys from a larger dictionary.

  Args:
    req_args: List of keys to extract
    all_kwargs: Dictionary with a superset of the required keys.
  Returns:
    req_args_dict: Dictionary with all the required arguments.
  """
  return {key: all_kwargs[key] for key in req_args}


def wrapped_general_conv(lhs, rhs, **kwargs):
  """Wrapped version of convolution that drop extra arguments.

  Args:
    lhs: First input to the convolution
    rhs: Second input to the convolution
    **kwargs: Dict with the parameters of the convolution with potentially
      some spurious parameters that the `lax.conv_general_dilated` would
      reject.
  Returns:
    out: Convolution output
  """
  req_arguments = ['window_strides', 'padding', 'lhs_dilation',
                   'rhs_dilation', 'dimension_numbers', 'feature_group_count',
                   'precision']
  lax_conv_params = collect_required_arguments(req_arguments, kwargs)
  return lax.conv_general_dilated(lhs, rhs, **lax_conv_params)


def simple_propagation(fn):
  """Create a wrapper function to ignore the context argument."""
  def wrapper(context, *args, **kwargs):
    del context
    return fn(*args, **kwargs)
  return wrapper



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of CrownIBP.

This corresponds to doing a forward pass based on IBP, followed by a bounding of
the output based on Crown bounds built using the IBP intermediate bounds.
"""
import functools
import jax

from jax import lax
from jax.interpreters import ad
import jax.numpy as jnp

from jax_verify.src import bound_propagation
from jax_verify.src import ibp
from jax_verify.src import utils
from jax_verify.src.fastlin import LinearExpression
import numpy as np


class CrownBackwardBound(bound_propagation.Bound):
  """Represent a pair of linear functions that bound the output of the network.

  The linear functions are stored as LinearExpressions. The linear coefficients
  correspond to activation at a given layer of the network and the offset are
  constant terms, of the shape of the network's output.

  Linear Coefficients will have the shape:
    batch_size x (output_shape) x (activation_shape)
  The offset will have the shape:
    batch_size x (output_shape)

  where output_shape and activation_shape can consist of multiple dimensions.
  """

  def __init__(self, lower_lin, upper_lin):
    self.lower_lin = lower_lin
    self.upper_lin = upper_lin

  @property
  def lower(self):
    raise NotImplementedError()

  @property
  def upper(self):
    raise NotImplementedError()

  def __add__(self, other):
    return self if other == 0 else CrownBackwardBound(
        self.lower_lin + other.lower_lin,
        self.upper_lin + other.upper_lin)

  def __radd__(self, other):
    return self if other == 0 else CrownBackwardBound(
        other.lower_lin + self.lower_lin,
        other.upper_lin + self.upper_lin)


def _broadcast_alltargets(arr, lincoeff):
  """Add unit dimensions to a Tensor to match the multiple targets.

  Args:
    arr: Tensor of shape batch_size x (activation_shape)
    lincoeff: Tensor of shape batch_size x (output_shape) x (activation_shape)
  Returns:
    reshaped_arr: `arr` input reshaped to have additional (output_shape) dims.
  """
  nb_dim_to_expand = lincoeff.ndim - arr.ndim
  target_shape = (arr.shape[0],) + (1,)*nb_dim_to_expand + arr.shape[1:]
  return jnp.reshape(arr, target_shape)


def concretize_backward_bound(backward_bound, act_bound):
  """Compute the value of a backward bound.

  Args:
    backward_bound: a CrownBackwardBound, representing linear functions of
     activations lower and upper bounding the output of the network.
    act_bound: Bound on the activation that the backward_bound is a function of.
  Returns:
    bound: A concretized bound
  """
  act_lower = _broadcast_alltargets(act_bound.lower,
                                    backward_bound.lower_lin.lin_coeffs)
  act_upper = _broadcast_alltargets(act_bound.upper,
                                    backward_bound.lower_lin.lin_coeffs)

  nb_dims_to_reduce = act_lower.ndim - backward_bound.lower_lin.offset.ndim
  dims_to_reduce = tuple(range(-nb_dims_to_reduce, 0))

  lower_lin = backward_bound.lower_lin
  upper_lin = backward_bound.upper_lin
  lower_bound = (lower_lin.offset
                 + jnp.sum(jnp.minimum(lower_lin.lin_coeffs, 0.) * act_upper
                           + jnp.maximum(lower_lin.lin_coeffs, 0.) * act_lower,
                           dims_to_reduce))
  upper_bound = (upper_lin.offset
                 + jnp.sum(jnp.maximum(upper_lin.lin_coeffs, 0.) * act_upper
                           + jnp.minimum(upper_lin.lin_coeffs, 0.) * act_lower,
                           dims_to_reduce))

  return ibp.IntervalBound(lower_bound, upper_bound)


def _crown_linear_op(lin_primitive, out_bound, *invals, **kwargs):
  """Backward propagation of LinearBounds through the primitive `lin_primitive`.

  This is achieved by piggybacking on the auto-differentiation code and relying
  on the fact that the backward propagation through linear layer is the
  transpose of the linear operation, which is the same as the operation done
  for gradient backpropagation.

  Args:
    lin_primitive: Jax primitive representing a bilinear operation, to go
      backward through.
    out_bound: CrownBackwardBound, linear function of the network outputs with
      regards to the output activation of this layer.
    *invals: input of the bound propagation in the forward pass
    **kwargs: Dict with the parameters of the linear operation.
  Returns:
    new_in_args: List of CrownBackwardBound
  """
  backward_primitive = ad.get_primitive_transpose(lin_primitive)

  to_backprop = jnp.concatenate([out_bound.lower_lin.lin_coeffs,
                                 out_bound.upper_lin.lin_coeffs], axis=1)
  nb_coeff_dim = (out_bound.lower_lin.lin_coeffs.ndim
                  - out_bound.lower_lin.offset.ndim)
  nb_output_dim = out_bound.lower_lin.offset.ndim - 1
  cts_in = to_backprop

  unwrapped_invals = []
  for inval in invals:
    if isinstance(inval, ibp.IntervalBound):
      # Create a fake input that would have matched the artificial construct
      # we defined as cts_in
      shape = cts_in.shape[:-nb_coeff_dim] + inval.shape[1:]
      unwrapped_invals.append(ad.UndefinedPrimal(jnp.zeros(shape)))
    elif isinstance(inval, jnp.ndarray):
      unwrapped_invals.append(inval)
    else:
      raise ValueError('Unexpected input for the crown-ibp'
                       f'primitive for {lin_primitive}.')

  vmap_outaxes = tuple(1 if isinstance(arg, ibp.IntervalBound) else None
                       for arg in invals)
  vmap_inaxes = (1,) + vmap_outaxes
  backward_op = functools.partial(backward_primitive, **kwargs)
  vmap_backward_op = backward_op
  for _ in range(nb_output_dim):
    # Vmap over all the dimensions that we need to pass through.
    vmap_backward_op = jax.vmap(vmap_backward_op, in_axes=vmap_inaxes,
                                out_axes=vmap_outaxes)
  cts_out = vmap_backward_op(cts_in, *unwrapped_invals)

  new_in_args = []
  for arg in cts_out:
    if arg is None:
      # This correspond to the input that was a constant, we don't want
      # to propagate anything there.
      new_in_args.append(arg)
    else:
      lower_lin_coeffs, upper_lin_coeffs = jnp.split(arg, 2, axis=1)
      new_in_args.append(CrownBackwardBound(
          LinearExpression(lower_lin_coeffs, out_bound.lower_lin.offset),
          LinearExpression(upper_lin_coeffs, out_bound.upper_lin.offset)))
  return new_in_args


def _crown_div(out_bound, lhs, rhs):
  """Backward propagation of LinearBounds through an addition.

  This is a linear operation only in the case where this is a division by a
  constant.

  Args:
    out_bound: CrownBackwardBound, linear function of network outputs bounds
      with regards to the results of the division.
    lhs: Numerator of the division.
    rhs: Denominator of the division.
  Returns:
    new_in_args: List of CrownBackwardBounds or Nones
  """
  if isinstance(rhs, bound_propagation.Bound):
    raise ValueError('Bound propagation through the denominator unsupported.')
  return _crown_linear_op(lax.div_p, out_bound, lhs, rhs)


def _crown_add(out_bound, lhs, rhs):
  """Backward propagation of LinearBounds through an addition.

  Args:
    out_bound: CrownBackwardBound, linear function of network outputs bounds
      with regards to the results of the addition.
    lhs: left input to the addition
    rhs: right input to the addition.
  Returns:
    new_in_args: List of CrownBackwardBounds or Nones
  """
  new_in_args = []
  for arg, other_arg in ((lhs, rhs), (rhs, lhs)):
    if isinstance(arg, ibp.IntervalBound):
      if isinstance(other_arg, ibp.IntervalBound):
        # If the two arguments are bounds, we should pass backward the output in
        # each of the branch.
        new_in_args.append(CrownBackwardBound(out_bound.lower_lin,
                                              out_bound.upper_lin))
      else:
        # If one arguments is a bound and the other is a constant, we should
        # pass backward the bound with a modified offset to the bound, and don't
        # need to propagate towards the constant.

        # Broadcast the `other` argument so that we can compute it's impact on
        # the final bound
        other_arg = _broadcast_alltargets(other_arg,
                                          out_bound.lower_lin.lin_coeffs)
        nb_dims_to_reduce = other_arg.ndim - out_bound.lower_lin.offset.ndim
        dims_to_reduce = tuple(range(-nb_dims_to_reduce, 0))

        lower_offset = jnp.sum(out_bound.lower_lin.lin_coeffs * other_arg,
                               dims_to_reduce)
        upper_offset = jnp.sum(out_bound.upper_lin.lin_coeffs * other_arg,
                               dims_to_reduce)

        new_in_arg = CrownBackwardBound(out_bound.lower_lin + lower_offset,
                                        out_bound.upper_lin + upper_offset)
        new_in_args.append(new_in_arg)
    elif isinstance(arg, jnp.ndarray):
      # This was simply a constant. We won't need to propagate further backwards
      new_in_args.append(None)
    else:
      raise ValueError('Unexpected input for _crown_add.')
  return new_in_args


def _crown_sub(out_bound, lhs, rhs):
  """Backward propagation of LinearBounds through a substraction.

  Args:
    out_bound: CrownBackwardBound, linear function of network outputs bounds
      with regards to the results of the addition.
    lhs: left input to the substraction
    rhs: right input to the substraction.
  Returns:
    new_in_args: List of CrownBackwardBounds or Nones
  """

  # We'll decompose this in the form of a multiplication of the right argument
  # by -1, followed by an addition.
  new_bound = _crown_add(out_bound, lhs, rhs)

  if isinstance(new_bound[1], CrownBackwardBound):
    to_neg = new_bound[1]
    new_bound[1] = CrownBackwardBound(
        LinearExpression(-to_neg.lower_lin.lin_coeffs, to_neg.lower_lin.offset),
        LinearExpression(-to_neg.upper_lin.lin_coeffs, to_neg.upper_lin.offset)
    )
  return new_bound


def _crown_max(out_bound, lhs, rhs):
  """Backward propagation of Linear Bounds through a ReLU.

  Args:
    out_bound: CrownBackwardBound, linear function of network outputs bounds
      with regards to the output of the ReLU
    lhs: left input to the max, inputs to the ReLU
    rhs: right input to the max, we assume this to be 0
  Returns:
    lhs_backbound: CrownBackwardBound, linear function of network outputs bounds
      with regards to the inputs of the ReLU.
    rhs_backbound: None, because we assume the second argument to be 0.
  """
  if not (isinstance(lhs, ibp.IntervalBound) and rhs == 0.):
    raise NotImplementedError('Only ReLU implemented for now.')

  relu_on = (lhs.lower >= 0.)
  relu_amb = jnp.logical_and(lhs.lower < 0., lhs.upper >= 0.)
  ub_slope = relu_on.astype(jnp.float32)
  ub_slope += jnp.where(relu_amb,
                        lhs.upper / jnp.maximum(lhs.upper - lhs.lower, 1e-12),
                        jnp.zeros_like(lhs.lower))
  ub_bias = jnp.where(relu_amb, - ub_slope * lhs.lower,
                      jnp.zeros_like(lhs.lower))
  # Crown Relu propagation.
  lb_slope = (ub_slope >= 0.5).astype(jnp.float32)
  lb_bias = jnp.zeros_like(ub_bias)

  lower_lin_coeffs = out_bound.lower_lin.lin_coeffs
  upper_lin_coeffs = out_bound.upper_lin.lin_coeffs

  ub_slope = _broadcast_alltargets(ub_slope, lower_lin_coeffs)
  lb_slope = _broadcast_alltargets(lb_slope, lower_lin_coeffs)

  new_lower_lin_coeffs = (jnp.minimum(lower_lin_coeffs, 0.) * ub_slope
                          + jnp.maximum(lower_lin_coeffs, 0.) * lb_slope)
  new_upper_lin_coeffs = (jnp.maximum(upper_lin_coeffs, 0.) * ub_slope
                          + jnp.minimum(upper_lin_coeffs, 0.) * lb_slope)

  bias_to_conc = ibp.IntervalBound(lb_bias, ub_bias)
  new_offset = concretize_backward_bound(out_bound, bias_to_conc)

  lhs_backbound = CrownBackwardBound(
      LinearExpression(new_lower_lin_coeffs, new_offset.lower),
      LinearExpression(new_upper_lin_coeffs, new_offset.upper))
  rhs_backbound = None

  return lhs_backbound, rhs_backbound


def _crown_reshape(out_bound, inp, new_sizes, dimensions):
  """Backward propagation of Linear Bounds through a reshape."""
  # Rather than parsing these, we will observe the shape of the inputs to deduce
  # what our target shape should be.
  del new_sizes, dimensions

  non_batch_inpshape = inp.shape[1:]

  # How many (output_shape) dimension is there.
  nb_target_dim = (out_bound.lower_lin.lin_coeffs.ndim
                   - out_bound.lower_lin.offset.ndim)
  batch_tgt_dims = nb_target_dim + 1

  lin_coeff_shape = out_bound.lower_lin.lin_coeffs.shape
  tgt_lin_coeff_shape = lin_coeff_shape[:batch_tgt_dims] + non_batch_inpshape
  new_lower_lin_coeffs = jnp.reshape(out_bound.lower_lin.lin_coeffs,
                                     tgt_lin_coeff_shape)
  new_upper_lin_coeffs = jnp.reshape(out_bound.upper_lin.lin_coeffs,
                                     tgt_lin_coeff_shape)
  inp_backbound = CrownBackwardBound(
      LinearExpression(new_lower_lin_coeffs, out_bound.lower_lin.offset),
      LinearExpression(new_upper_lin_coeffs, out_bound.upper_lin.offset)
  )
  new_sizes_backbound = None
  return inp_backbound, new_sizes_backbound


def crownibp_bound_propagation(function, bounds):
  """Performs Crown-IBP as described in https://arxiv.org/abs/1906.06316.

  We first perform IBP to obtain intermediate bounds and then propagate linear
  bounds backwards.

  Args:
    function: Function performing computation to obtain bounds for. Takes as
      only argument the network inputs.
    bounds: jax_verify.IntervalBounds, bounds on the inputs of the function.
  Returns:
    output_bound: Bounds on the output of the function obtained by Crown-IBP
  """
  ibp_bound, graph = bound_propagation.bound_propagation(
      ibp.bound_transform, function, bounds)

  # Define the initial bound to propagate backward
  assert hasattr(ibp_bound, 'shape'), (
      'crownibp_bound_propagation requires `function` to output a single array '
      'as opposed to an arbitrary pytree')
  batch_size = ibp_bound.shape[0]
  act_shape = ibp_bound.shape[1:]

  nb_act = np.prod(act_shape)
  identity_lin_coeffs = jnp.reshape(jnp.eye(nb_act), act_shape + act_shape)
  initial_lin_coeffs = jnp.repeat(jnp.expand_dims(identity_lin_coeffs, 0),
                                  batch_size, axis=0)
  initial_offsets = jnp.zeros_like(ibp_bound.lower)

  initial_backward_bound = CrownBackwardBound(
      LinearExpression(initial_lin_coeffs, initial_offsets),
      LinearExpression(initial_lin_coeffs, initial_offsets))

  input_fun, = graph.backward_propagation(
      _primitive_transform, sum, initial_backward_bound)

  return concretize_backward_bound(input_fun, bounds)


_primitive_transform = {
    lax.add_p: _crown_add,
    lax.sub_p: _crown_sub,
    lax.reshape_p: _crown_reshape,
    lax.dot_general_p: functools.partial(_crown_linear_op, lax.dot_general_p),
    lax.mul_p: functools.partial(_crown_linear_op, lax.mul_p),
    lax.div_p: _crown_div,
    lax.conv_general_dilated_p: functools.partial(
        _crown_linear_op, lax.conv_general_dilated_p),
    lax.max_p: _crown_max
}

_primitive_transform = {key: utils.simple_propagation(prop_fun)
                        for key, prop_fun in _primitive_transform.items()}



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Methods for solving relaxation generated by relaxation.py.

This file mainly calls out to relaxation.RelaxationSolvers defined in other
files, and provides a higher-level interface than using relaxation.py directly.
"""

from jax_verify.src import bound_propagation
from jax_verify.src import cvxpy_relaxation_solver
from jax_verify.src import relaxation


def solve_planet_relaxation(
    logits_fn, initial_bounds, boundprop_transform,
    objective, objective_bias, index,
    solver=cvxpy_relaxation_solver.CvxpySolver):
  """Solves the "Planet" (Ehlers 17) or "triangle" relaxation.

  The general approach is to use jax_verify to generate constraints, which can
  then be passed to generic solvers. Note that using CVXPY will incur a large
  overhead when defining the LP, because we define all constraints element-wise,
  to avoid representing convolutional layers as a single matrix multiplication,
  which would be inefficient. In CVXPY, defining large numbers of constraints is
  slow.

  Args:
    logits_fn: Mapping from inputs (batch_size x input_size) -> (batch_size,
      num_classes)
    initial_bounds: `IntervalBound` with initial bounds on inputs,
      with lower and upper bounds of dimension (batch_size x input_size).
    boundprop_transform: bound_propagation.BoundTransform instance, such as
      `jax_verify.ibp_transform`. Used to pre-compute interval bounds for
      intermediate activations used in defining the Planet relaxation.
    objective: Objective to optimize, given as an array of coefficients to be
      applied to the output of logits_fn defining the objective to minimize
    objective_bias: Bias to add to objective
    index: Index in the batch for which to solve the relaxation
    solver: A relaxation.RelaxationSolver, which specifies the backend to solve
      the resulting LP.
  Returns:
    val: The optimal value from the relaxation
    status: The status of the relaxation solver
  """
  relaxation_transform = relaxation.RelaxationTransform(boundprop_transform)
  variable, graph = bound_propagation.bound_propagation(
      relaxation_transform, logits_fn, initial_bounds)
  value, status = relaxation.solve_relaxation(
      solver, objective, objective_bias, variable, graph.env,
      index=index, time_limit=None)
  return value, status



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides methods to simplify the JaxPR computation graph.
"""
import jax
import jax.numpy as jnp


def activation_detector(graph):
  """Attempts to simplify the graph by identifying specific activations.

  This is done by recognizing part of the graph and fusing them into synthetic
  primitives. At the moment, only Softplus is detected.

  Args:
    graph: Unprocessed JaxPR graph.
  Returns:
    simple_graph: Potentially modified JaxPR graph.
  """
  new_eqns = []
  eqn_idx = 0
  while eqn_idx < len(graph.eqns):
    if _is_softplus(graph, eqn_idx):
      # In JaxPR, softplus is a call to logaddexp(x, 0) but our restricted
      # primitive only takes x as input.
      softplus_outvars = graph.eqns[eqn_idx + 9].outvars
      softplus_invars = graph.eqns[eqn_idx].invars[:1]
      softplus_eqn = jax.core.new_jaxpr_eqn(softplus_invars, softplus_outvars,
                                            softplus_p, {})
      new_eqns.append(softplus_eqn)
      eqn_idx += 10
    elif graph.eqns[eqn_idx] == jax.lax.convert_element_type_p:
      # Skip this, this should only be consumed by the softplus
      eqn_idx += 1
    else:
      new_eqns.append(graph.eqns[eqn_idx])
      eqn_idx += 1

  simple_graph = jax.core.Jaxpr(graph.constvars, graph.invars,
                                graph.outvars, new_eqns)
  return simple_graph


def _is_softplus(graph, eqn_idx):
  """Check if the the position `eqn_idx` is the start of a softplus sequence."""
  if eqn_idx + 10 > len(graph.eqns):
    # Softplus takes 10 primitive to implement
    return False

  # The last operations should always be the same.
  valid_primitives = (
      (graph.eqns[eqn_idx + 6].primitive == jax.lax.exp_p) and
      (graph.eqns[eqn_idx + 7].primitive == jax.lax.log1p_p) and
      (graph.eqns[eqn_idx + 8].primitive == jax.lax.add_p) and
      (graph.eqns[eqn_idx + 9].primitive == jax.lax.select_p))
  if not valid_primitives:
    return False

  # Check that this is wired correctly to correspond to a softplus.
  # There is a NaN-check
  for pos in range(9):
    if graph.eqns[eqn_idx + pos].primitive == jax.lax.ne_p:
      ne_eqn = graph.eqns[eqn_idx + pos]
      # It's a NaN check if the two inputs compared for equality are the same.
      if ne_eqn.invars[0] != ne_eqn.invars[1]:
        return False
  # Output of the exponential goes to log1p
  if graph.eqns[eqn_idx + 6].outvars[0] != graph.eqns[eqn_idx + 7].invars[0]:
    return False
  # The select is based on the NaN check
  if graph.eqns[eqn_idx + 9].invars[0] != ne_eqn.outvars[0]:
    return False
  # The Non-Nan selected input is the result of the add
  if graph.eqns[eqn_idx + 9].invars[2] != graph.eqns[eqn_idx + 8].outvars[0]:
    return False

  return True


softplus_p = jax.lax.standard_naryop([{jnp.floating}], 'Softplus')
softplus_p.def_impl(jax.nn.softplus)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
# pylint: disable=invalid-name
"""Small helper functions."""

import collections
import copy
import enum
import functools

import jax
import jax.lax as lax
import jax.numpy as jnp
import numpy as np
import optax
import tree

IntBound = collections.namedtuple(
    'IntervalBound', ['lb', 'ub', 'lb_pre', 'ub_pre'])
_AdvRobustnessVerifInstance = collections.namedtuple(
    'VerifInstance',
    ['params', 'params_full', 'input_bounds', 'bounds', 'obj', 'obj_orig',
     'const', 'type'])
_VerifInstance = _AdvRobustnessVerifInstance  # alias
_SdpDualVerifInstance = collections.namedtuple(
    'SdpDualVerifInstance',
    ['bounds', 'make_inner_lagrangian', 'dual_shapes', 'dual_types'])


class SdpDualVerifInstance(_SdpDualVerifInstance):
  """A namedtuple specifying a verification instance for the dual SDP solver.

  Fields:
    * bounds: A list of bounds on post-activations at each layer
    * make_inner_lagrangian: A function which takes ``dual_vars`` as input, and
      returns another function, the inner lagrangian, which evaluates
      Lagrangian(x, dual_vars) for any value ``x`` (the set of activations).
    * dual_types: A pytree matching dual_vars specifying which dual_vars
      should be non-negative.
    * dual_shapes: A pytree matching dual_vars specifying shape of each var.
  """


default_conv_transpose = functools.partial(
    lax.conv_transpose, dimension_numbers=('NHWC', 'HWIO', 'NHWC'))

################## Networks ######################


def mlp_layer_sizes(params):
  # Dimension of input to each layer, and output of final layer
  layer_sizes = [int(w.shape[0]) for (w, b) in params]
  fin_w, _ = params[-1]
  layer_sizes.append(int(fin_w.shape[1]))
  return layer_sizes


def nn_layer_sizes(params):
  """Compute MLP sizes of the inputs/outputs for individual layers."""
  assert not any([isinstance(x, dict) for x in params]), 'MLP only'
  return mlp_layer_sizes(params)


def layer_sizes_from_bounds(bounds):
  assert all([b.lb.shape[0] == 1 for b in bounds])
  layer_sizes = [b.lb.shape[1:] for b in bounds]
  layer_sizes = [s[0] if len(s) == 1 else s for s in layer_sizes]
  return layer_sizes


def predict_mlp(params, inputs):
  for W, b in params[:-1]:
    outputs = jnp.dot(inputs, W) + b
    inputs = jnp.maximum(outputs, 0)
  W, b = params[-1]
  return jnp.dot(inputs, W) + b


def fwd(inputs, layer_params):
  """JAX forward pass of Linear, Conv, or ConvTranspose."""
  if isinstance(layer_params, dict):
    # Conv/ConvTranspose: Reshape input if necessary:
    if len(inputs.shape) < 4:
      w = h = int(np.sqrt(inputs.shape[-1]/layer_params['n_cin']))
      inputs = inputs.reshape(inputs.shape[0], h, w, layer_params['n_cin'])
    W, b = layer_params['W'], np.reshape(layer_params['b'], [1, 1, 1, -1])
    if 'transpose' in layer_params and layer_params['transpose']:
      # ConvTranspose
      return default_conv_transpose(
          inputs, W, (layer_params['stride'], layer_params['stride']),
          layer_params['padding']) + b
    else:
      # Normal Conv2D
      dn = lax.conv_dimension_numbers(inputs.shape, W.shape,
                                      ('NHWC', 'HWIO', 'NHWC'))
      return lax.conv_general_dilated(
          inputs, W, (layer_params['stride'], layer_params['stride']),
          layer_params['padding'], (1, 1), (1, 1), dn) + b
  elif isinstance(layer_params, tuple):
    # Linear fully-connected layer
    # TODO: Figure out why we were dropping batch dim before here
    inputs = (inputs.reshape(inputs.shape[0], -1)
              if len(inputs.shape) == 4 else inputs)
    (W, b) = layer_params
    return jnp.dot(inputs, W) + b
  else:
    raise NotImplementedError('Unknown layer')


def predict_cnn(params, inputs, include_preactivations=False):
  """Forward pass for a CNN given parameters.

  Args:
    params: Parameters for the CNN. See make_cnn_params for syntax.
    inputs: Inputs to CNN.
    include_preactivations: bool. If True, also return pre-activations after
      each matmul layer.
  Returns:
    act: Output from forward pass through CNN.
    (Optional) layer_acts: Post-relu activation at each layer
  """
  act = inputs
  layer_preacts = []
  for counter, layer_params in enumerate(params):
    act = fwd(act, layer_params)
    layer_preacts.append(act)
    if counter < len(params) - 1:
      # no relu on final layer
      act = jnp.maximum(act, 0)
  return act if not include_preactivations else (act, layer_preacts)


def get_network_activs(params, x):
  assert len(x.shape) == 1, 'x should not have batch dim'
  activs = [x]
  for W, b in params:
    x = jnp.matmul(x, W) + b
    x = jnp.maximum(x, 0.)
    activs.append(x)
  return activs


def get_layer_params(fun_to_extract, example_input):
  """Extract the parameters from a network specified as a function.

  Args:
    fun_to_extract: Function implementing a simple MLP/CNN network composed
      of alternating linear layers and ReLU activation functions.
    example_input: Example of input to `fun_to_extract`.
  Returns:
    params: Parameters for the CNN/MLP, as taken by `predict_cnn` or
      `predict_mlp`
  """
  jaxpr_maker = jax.make_jaxpr(fun_to_extract)
  parsed = jaxpr_maker(example_input)

  layers = []
  next_is_relu = False
  scaling = None
  centering = None

  jax_parameters = {var: param_vals
                    for var, param_vals in zip(parsed.jaxpr.constvars,
                                               parsed.literals)}
  def _get_const_input_arg(eqn):
    if eqn.invars[0] in jax_parameters:
      return eqn.invars[0]
    elif eqn.invars[1] in jax_parameters:
      return eqn.invars[1]
    else:
      raise ValueError('None of the primitive\'s input is a weight tensor.')

  for eqn in parsed.jaxpr.eqns:
    if eqn.primitive == lax.reshape_p:
      if eqn.invars[0] in jax_parameters:
        # If this is a reshaping of a constant / the input of the reshape will
        # be in the jax_parameters dict. We can treat the reshaped constant as a
        # another constant.
        out_var = eqn.outvars[0]
        inps = [jax_parameters[eqn.invars[0]]]
        jax_parameters[out_var] = eqn.primitive.bind(*inps, **eqn.params)
      else:
        # If it is a reshape on the pass of the network forward propagation, we
        # can ignore it as the forward evaluation code in sdp_verify handles the
        # reshaping itself
        continue
    elif eqn.primitive in (lax.dot_general_p,
                           lax.conv_general_dilated_p):
      if next_is_relu:
        raise ValueError('Unsupported architecture. Only supported networks are'
                         ' alternance of linear (convolutional/fully connected)'
                         ' and ReLU layers.')
      # Find the input which is a parameter.
      param_input = _get_const_input_arg(eqn)
      weight_params = jax_parameters[param_input]

      if eqn.primitive == lax.dot_general_p:
        bias_shape = weight_params.shape[1]
        scaling_shape = (-1, 1)
      else:
        bias_shape = (1, 1, 1, weight_params.shape[-1])
        # Based on the code in `fwd`, the dimension for the input channel is the
        # third one.
        scaling_shape = (1, 1, -1, 1)

      # Define the bias of the network, potentially incorporating existing
      # preprocessing steps
      bias = jnp.zeros(bias_shape)
      if centering is not None:
        inp_bias = jnp.zeros_like(example_input) + centering
        equivalent_bias = eqn.primitive.bind(inp_bias, weight_params,
                                             **eqn.params)
        bias += equivalent_bias
        centering = None
      if scaling is not None:
        scaling = jnp.reshape(scaling, scaling_shape)
        weight_params = weight_params * scaling
        scaling = None

      if eqn.primitive == lax.dot_general_p:
        layers.append((weight_params, bias))
      else:
        # The eval function of sdp_verify only handle stride equal
        # that are the same in all directions
        strides = eqn.params['window_strides']
        if not all(elt == strides[0] for elt in strides):
          raise ValueError('Different spatial strides unsupported.')
        # The forward code expect the bias to only correspond to one column.
        bias = jnp.reshape(bias[0, 0, 0, :], (1, 1, 1, -1))
        layers.append({
            'W': weight_params,
            'b': bias,
            'stride': strides[0],
            'padding': eqn.params['padding'],
        })
      next_is_relu = True
    elif eqn.primitive == lax.add_p:
      param_input = _get_const_input_arg(eqn)
      bias_params = jax_parameters[param_input]
      if not next_is_relu:
        raise ValueError('Unsupported architecture. Only supported networks are'
                         ' alternance of linear (convolutional/fully connected)'
                         ' and ReLU layers.')
      # This is an addition after a linear layer. Just fold it into the bias
      # of the previous layer.
      if isinstance(layers[-1], tuple):
        # Remove the last linear layer and include a version of it that includes
        # the bias term.
        weight, bias = layers.pop()
        layers.append((weight, bias + bias_params))
      else:
        layers[-1]['b'] = layers[-1]['b'] + bias_params
      # No need to update `next_is_relu` because it remains True
    elif eqn.primitive == lax.sub_p:
      if layers:
        # We handle this only in the case of preprocessing at the beginning of
        # the network.
        raise ValueError('Unsupported operation. sub is only supported as a'
                         'centering of the networks inputs.')
      # This appears potentially at the beginning of the network, during the
      # preprocessing of the inputs.
      centering = centering or 0.
      centering -= jax_parameters[_get_const_input_arg(eqn)]
    elif eqn.primitive == lax.div_p:
      if layers:
        # We handle this only in the case of preprocessing at the beginning of
        # the network.
        raise ValueError('Unsupported operation. div is only supported as a'
                         'rescaling of the networks inputs.')
      divide_scaling = jax_parameters[_get_const_input_arg(eqn)]
      # This appears during the preprocessing of the inputs of the networks.
      scaling = scaling or 1.0
      scaling /= divide_scaling
      # Rescale the centering if there is one.
      if centering is not None:
        centering /= divide_scaling
    elif eqn.primitive == lax.max_p:
      if ((not next_is_relu)
          or (not isinstance(eqn.invars[1], jax.core.Literal))
          or (eqn.invars[1].val != 0.0)):
        raise ValueError('Unsupported architecture. Only supported networks are'
                         ' alternance of linear (convolutional/fully connected)'
                         ' and ReLU layers.')
      # The ReLU are not denoted in the parameters dictionaries so no need to
      # add anything to the layers list.
      next_is_relu = False
    elif eqn.primitive == lax.broadcast_in_dim_p:
      # There might be broadcast of bias, we'll just store the original bias
      # where the broadcasted one should be.
      jax_parameters[eqn.outvars[0]] = jax_parameters[eqn.invars[0]]
    else:
      raise ValueError(f'Unsupported primitive {eqn.primitive}. The only '
                       'supported networks are alternance of linear '
                       '(convolutional / fully connected) and ReLU layers.')
  return layers


################## Bound prop ####################


def boundprop(params, bounds_in):
  """Compute IntervalBound for each layer."""
  layer_bounds = [bounds_in]
  for layer_params in params:
    lb_old = layer_bounds[-1].lb
    ub_old = layer_bounds[-1].ub

    if isinstance(layer_params, dict):
      center = (lb_old + ub_old) / 2.
      radius = (ub_old - lb_old) / 2.
      rad_layer_params = copy.deepcopy(layer_params)
      rad_layer_params['W'] = jnp.abs(layer_params['W'])
      rad_layer_params['b'] = jnp.zeros_like(layer_params['b'])
      act_c = fwd(center, layer_params)
      act_r = fwd(radius, rad_layer_params)
      lb = act_c - act_r
      ub = act_c + act_r
    else:
      if len(lb_old.shape) == 4:
        lb_old = lb_old.reshape((lb_old.shape[0], -1))
        ub_old = ub_old.reshape((ub_old.shape[0], -1))
      # TODO: Post-neurips. This logic can also be shared with above.
      W, b = layer_params

      lb = (jnp.matmul(lb_old, jnp.maximum(W, 0.)) +
            jnp.matmul(ub_old, jnp.minimum(W, 0.)) +
            b)
      ub = (jnp.matmul(ub_old, jnp.maximum(W, 0.)) +
            jnp.matmul(lb_old, jnp.minimum(W, 0.)) +
            b)

    layer_bounds.append(IntBound(lb_pre=lb,
                                 ub_pre=ub,
                                 lb=jnp.maximum(lb, 0.),
                                 ub=jnp.maximum(ub, 0.)))

  return layer_bounds


def init_bound(x, epsilon, input_bounds=(0., 1.), add_batch_dim=True):
  x = np.expand_dims(x, axis=0) if add_batch_dim else x
  lb_init = np.maximum(input_bounds[0], x - epsilon)
  ub_init = np.minimum(input_bounds[1], x + epsilon)
  return IntBound(lb=lb_init, ub=ub_init, lb_pre=None, ub_pre=None)


def ibp_bound_elided(verif_instance):
  assert len(verif_instance.bounds) == len(verif_instance.params) + 2
  obj, obj_const, final_bound = (
      verif_instance.obj, verif_instance.const, verif_instance.bounds[-2])
  ub = final_bound.ub.reshape(final_bound.ub.shape[0], -1)
  lb = final_bound.lb.reshape(final_bound.lb.shape[0], -1)
  obj_val = np.sum(np.maximum(obj, 0.) * ub +
                   np.minimum(obj, 0.) * lb)
  return float(obj_val + obj_const)


def ibp_bound_nonelided(verif_instance):
  assert len(verif_instance.bounds) == len(verif_instance.params) + 2
  obj_orig = verif_instance.obj_orig
  final_bound = verif_instance.bounds[-1]
  batch_size = final_bound.ub_pre.shape[0]
  ub = final_bound.ub_pre.reshape(batch_size, -1)
  lb = final_bound.lb_pre.reshape(batch_size, -1)
  obj_val = np.sum(np.maximum(obj_orig, 0.) * ub +
                   np.minimum(obj_orig, 0.) * lb)
  return float(obj_val)


################## Solver-agnostic Verification Instances ####################


class VerifInstanceTypes(enum.Enum):
  # `params` represent a network of repeated relu(Wx+b)
  # The final output also includes a relu activation, and `obj` composes
  # the final layer weights with the original objective
  MLP_ELIDED = 'mlp_elided'
  CNN_ELIDED = 'cnn_elided'


def make_relu_robust_verif_instance(
    params, bounds=None, target_label=1, label=2, input_bounds=None):
  """Make VerifInstance from network weights and input.

  Args:
    params: list of pairs of array-like objects [(W, b)], the weights and biases
      of a multi-layer perceptron.
    bounds: None, or a list of IntBound objects, of length len(params) + 1.
      The interval bounds for each layer.
    target_label: int, the adversary target
    label: int, the true label
    input_bounds: None, pair of floats, or pair of vectors with length matching
      input dimension. The image bounds e.g. (0, 1) or (0, 255).

  Returns:
    verif_instance: a VerifInstance object
  """
  assert bounds is None or len(bounds) == len(params) + 1
  if bounds is None:
    assert not any([isinstance(x, dict) for x in params])
    layer_sizes = mlp_layer_sizes(params)
  else:
    layer_sizes = layer_sizes_from_bounds(bounds)
  # Adversary maximizes objective - large when logit(target) > logit(label)
  if any([isinstance(x, dict) for x in params]):
    verif_type = VerifInstanceTypes.CNN_ELIDED
  else:
    verif_type = VerifInstanceTypes.MLP_ELIDED
  label_onehot = np.eye(int(layer_sizes[-1]))[label]
  target_onehot = np.eye(int(layer_sizes[-1]))[target_label]
  obj_orig = target_onehot - label_onehot
  w_fin, b_fin = params[-1]
  obj_bp = np.matmul(w_fin, obj_orig)
  const = np.vdot(obj_orig, b_fin)
  obj = np.reshape(obj_bp, (1, obj_bp.size))
  return _VerifInstance(
      params=params[:-1],
      params_full=params,
      input_bounds=input_bounds,
      bounds=bounds,
      obj=obj,
      obj_orig=obj_orig,
      const=const,
      type=verif_type)

make_nn_verif_instance = make_relu_robust_verif_instance  # alias

################## SDP Verification Instances ####################

# Dual variables correspond to:
#   lam: ReLU quadratic constraint: z^2 = z*(Wx)
#   nu: IBP quadratic constraint: x^2 <= (l+u)*x - l*u
#   nu_quad: IBP quadratic matrix constraint: (x_i - l_i)(x_j - u_j) <= 0
#   muminus: x'>=0
#   muminus2: Triangle linear Relu relaxation - u(Wx+b) - ul - (u-l)x' >= 0
#       where l = min(l, 0), u = max(u, 0)
#   muplus: x'>=Wx+b
DualVar = collections.namedtuple(
    'DualVar', ['lam', 'nu', 'nu_quad', 'muminus', 'muplus', 'muminus2'])
DualVarFin = collections.namedtuple('DualVarFin', ['nu', 'nu_quad'])
DEFAULT_DISABLED_DUAL_VARS = ('nu_quad', 'muminus2')
NECESSARY_DUAL_VARS = ('lam', 'muplus', 'muminus')


class DualVarTypes(enum.Enum):
  EQUALITY = 'equality'
  INEQUALITY = 'inequality'


def make_relu_network_lagrangian(dual_vars, params, bounds, obj):
  """Returns a function that computes the Lagrangian for a ReLU network.

  This function assumes `params` represent a feedforward ReLU network i.e.
  x_{i+1} = relu(W_i x_i + b_i). It defines the Lagrangian by applying the
  objective `obj` to the final layer activations, and encoding the Lagrangian
  terms for each of the constraints defining the ReLU network. It then returns
  this function.

  Args:
    dual_vars: A length L+1 list of dual variables at each layer
    params: A length L list of (W, b) pairs, elided network weights
    bounds: A length L+1 list of `IntBound`s, elided bounds at each layer
    obj: function, taking final layer activations as input

  Returns:
    Function that computes Lagrangian L(x) with fixed `dual_vars`.
  """
  layer_sizes = layer_sizes_from_bounds(bounds)

  def lagrangian(xs_list):
    """Computes Lagrangian L(x) with fixed `dual_vars`."""
    assert all([x.shape[0] == 1 for x in xs_list]), 'no batch mode support'

    lag = obj(xs_list[-1])
    for i in range(len(layer_sizes)):
      if i < len(params):
        y = fwd(xs_list[i], params[i])
        # Lagrangian for constraint x' * x' = x' * (Wx+b) where x'=ReLU(Wx+b)
        lag += (jnp.sum(dual_vars[i].lam * xs_list[i + 1] *
                        (y - xs_list[i + 1])))
        # Lagrangian for the constraint x'>=Wx+b
        lag += jnp.sum(dual_vars[i].muplus * (xs_list[i + 1] - y))
        if dual_vars[i].muminus2.shape:
          # Lagrangian for u(Wx+b) - ul - (u-l)x' >= 0, where
          # l = min(l, 0) and u = max(u, 0)
          raise NotImplementedError('dropped support for muminus2')

        # Lagrangian for the constraint x'>=0
        lag += jnp.sum(dual_vars[i].muminus * xs_list[i + 1])

      # Lagrangian for IBP constraint (x-l)(x-u) <= 0
      if dual_vars[i].nu.shape:
        lag += -jnp.sum(dual_vars[i].nu *
                        (xs_list[i] - bounds[i].lb) *(xs_list[i] - bounds[i].ub)
                        )
      if dual_vars[i].nu_quad.shape:
        # IBP quadratic matrix constraint: (x_i - l_i)(x_j - u_j) <= 0
        lag += -jnp.sum(dual_vars[i].nu_quad *
                        jnp.matmul((xs_list[i]-bounds[i].lb).T,
                                   xs_list[i]-bounds[i].ub))
    return lag
  return lagrangian


def relu_robustness_verif_instance_to_sdp(verif_instance):
  """Convert solver-agnostic VerifInstance to SdpDualVerifInstance."""
  assert verif_instance.type in [
      VerifInstanceTypes.MLP_ELIDED, VerifInstanceTypes.CNN_ELIDED]
  elided_bounds = verif_instance.bounds[:-1]
  dual_shapes, dual_types = get_dual_shapes_and_types(elided_bounds)
  def obj(x_final):
    out = jnp.sum(x_final * jnp.reshape(verif_instance.obj, x_final.shape))
    return out + verif_instance.const
  def make_inner_lagrangian(dual_vars):
    return make_relu_network_lagrangian(
        dual_vars, verif_instance.params, elided_bounds, obj)
  return SdpDualVerifInstance(
      make_inner_lagrangian=make_inner_lagrangian,
      bounds=elided_bounds,
      dual_shapes=dual_shapes,
      dual_types=dual_types)


def make_sdp_verif_instance(verif_instance):
  if isinstance(verif_instance, _AdvRobustnessVerifInstance):
    return relu_robustness_verif_instance_to_sdp(verif_instance)
  else:
    raise NotImplementedError('unrecognized verif_instance type')


def make_vae_sdp_verif_instance(params, data_x, bounds):
  """Make SdpDualVerifInstance for VAE reconstruction error spec."""
  elided_params = params[:-1]
  elided_bounds = bounds[:-1]
  dual_shapes, dual_types = get_dual_shapes_and_types(elided_bounds)
  def recon_loss(x_final):
    x_hat = predict_cnn(params[-1:], x_final).reshape(1, -1)
    return jnp.sum(jnp.square(data_x.reshape(x_hat.shape) - x_hat))
  def make_inner_lagrangian(dual_vars):
    return make_relu_network_lagrangian(
        dual_vars, elided_params, elided_bounds, recon_loss)
  return SdpDualVerifInstance(
      make_inner_lagrangian=make_inner_lagrangian,
      bounds=elided_bounds,
      dual_shapes=dual_shapes,
      dual_types=dual_types)


def make_vae_semantic_spec_params(x, vae_params, classifier_params):
  """Defines network f(z_noise) = classifier(reconstruct(x, z_noise))."""
  # Setup - encoder fwd pass
  encoder_params, decoder_params = vae_params
  encoder_mu_params = encoder_params[:-1]
  encoder_sigmasq_params = encoder_params[:-2] + [encoder_params[-1]]
  mu_z = predict_cnn(encoder_mu_params, x)
  log_sigmasq_z = predict_cnn(encoder_sigmasq_params, x)
  sigmasq_z = jnp.exp(log_sigmasq_z)
  # Combine the reparameterization with the first decoder layer
  # z0 = mu + sigma * z
  # z1 = jnp.dot(z0, W) + b
  #    = jnp.dot(mu + sigma * z, W) + b
  #    = jnp.dot(z, sigma * W) + [b + jnp.dot(mu, W)]
  assert isinstance(decoder_params[0], tuple)
  W0_orig, b0_orig = decoder_params[0]
  W0 = W0_orig * jnp.reshape(jnp.sqrt(sigmasq_z), (-1, 1))
  b0 = b0_orig + jnp.dot(mu_z, W0_orig)

  # Now the network is just concatenation of modified decoder + classifier
  # This is also applying a Relu to decoder output, but that's fine
  combined_params = [(W0, b0)] + decoder_params[1:] + classifier_params
  return combined_params


def get_dual_shapes_and_types(bounds_elided):
  """Get shapes and types of dual vars."""
  dual_shapes = []
  dual_types = []
  layer_sizes = layer_sizes_from_bounds(bounds_elided)
  for it in range(len(layer_sizes)):
    m = layer_sizes[it]
    m = [m] if isinstance(m, int) else list(m)
    if it < len(layer_sizes)-1:
      n = layer_sizes[it + 1]
      n = [n] if isinstance(n, int) else list(n)
      shapes = {
          'lam': [1] + n,
          'nu': [1] + m,
          'muminus': [1] + n,
          'muplus': [1] + n,
          'nu_quad': [], 'muminus2': [],
      }
      types = {
          'lam': DualVarTypes.EQUALITY,
          'nu': DualVarTypes.INEQUALITY,
          'muminus': DualVarTypes.INEQUALITY,
          'muplus': DualVarTypes.INEQUALITY,
          'nu_quad': DualVarTypes.INEQUALITY,
          'muminus2': DualVarTypes.INEQUALITY,
      }
      dual_shapes.append(DualVar(**{
          k: np.array(s) for k, s in shapes.items()}))
      dual_types.append(DualVar(**types))
    else:
      shapes = {'nu': [1] + m, 'nu_quad': []}
      types = {'nu': DualVarTypes.INEQUALITY,
               'nu_quad': DualVarTypes.INEQUALITY}
      dual_shapes.append(DualVarFin(**{
          k: np.array(s) for k, s in shapes.items()}))
      dual_types.append(DualVarFin(**types))

  # Add kappa
  N = sum([np.prod(np.array(i)) for i in layer_sizes])
  dual_shapes.append(np.array([1, N+1]))
  dual_types.append(DualVarTypes.INEQUALITY)
  return dual_shapes, dual_types


################### Image Preprocessing #######################


def preprocess_cifar(image, inception_preprocess=False, perturbation=False):
  """Proprocess images and perturbations."""
  if inception_preprocess:
    # Use 2x - 1 to get [-1, 1]-scaled images
    rescaled_devs = 0.5
    rescaled_means = 0.5
  else:
    rescaled_means = np.array([125.3, 123.0, 113.9]) / 255
    rescaled_devs = np.array([63.0, 62.1, 66.7]) / 255
  if perturbation:
    return image / rescaled_devs
  else:
    return (image - rescaled_means) / rescaled_devs


def preprocessed_cifar_eps_and_input_bounds(
    shape=(32, 32, 3), epsilon=2/255, inception_preprocess=False):
  """Get `epsilon` and `input_bounds`."""
  preprocess = functools.partial(
      preprocess_cifar, inception_preprocess=inception_preprocess)
  epsilon = preprocess(np.ones(shape)* epsilon, perturbation=True)
  input_bounds = (preprocess(np.zeros(shape)), preprocess(np.ones(shape)))
  return epsilon, input_bounds


################### Attacks #######################


def adv_objective(model_fn, x, label, target_label):
  logits = model_fn(x)
  if len(logits.shape) == 2:
    logits = logits[0]
  loss = logits[target_label] - logits[label]
  return jnp.sum(loss)


def fgsm_single(model_fn, x, label, target_label, epsilon, num_steps,
                step_size, input_bounds=(0., 1.)):
  """Same interface as l.d.r.adversarial.attacks, but no batch dim on x."""
  adv_loss = lambda *args, **kwargs: -adv_objective(*args, **kwargs)
  adv_loss_x = lambda x: adv_loss(model_fn, x, label, target_label)
  return pgd(adv_loss_x, x, epsilon, num_steps, step_size,
             input_bounds=input_bounds)


def untargeted_margin_loss(logits, labels):
  """Minimized by decreasing true score, and increasing second highest."""
  batch_size = logits.shape[0]
  num_classes = logits.shape[-1]
  label_logits = logits[jnp.arange(batch_size), labels]
  logit_mask = jax.nn.one_hot(labels, num_classes)
  inf = 1e5
  highest_logits = jnp.max(logits - inf * logit_mask, axis=-1)
  return label_logits - highest_logits


def pgd_default(model_fn, x, label, epsilon, num_steps, step_size,
                input_bounds=(0., 1.)):
  assert x.shape[0] == label.shape[0]
  adv_loss_x = lambda x: jnp.sum(untargeted_margin_loss(model_fn(x), label))
  return pgd(adv_loss_x, x, epsilon, num_steps, step_size,
             input_bounds=input_bounds)


def pgd(adv_loss, x_init, epsilon, num_steps, step_size, input_bounds=(0., 1.)):
  grad_adv_loss = jax.grad(adv_loss)
  x = x_init
  for _ in range(num_steps):
    grad_x = grad_adv_loss(x)
    x -= jnp.sign(grad_x) * step_size
    x = jnp.clip(x, x_init - epsilon, x_init + epsilon)
    x = jnp.clip(x, input_bounds[0], input_bounds[1])
  return x


################ Optimizers ##################


def scale_by_variable_opt(multipliers):
  """Custom learning rates for different variables.

  Args:
    multipliers: a pytree, with the same structure as `params`. Each leaf can
      be either a float, or an array shape-compatible with the corresponding
      `params` element. These multiply the learning rate for each leaf.

  Returns:
    optax.GradientTransformation optimizer
  """

  def init_fn(params):
    params_struct = jax.tree_map(lambda _: None, params)
    multipliers_struct = jax.tree_map(lambda _: None, multipliers)
    assert params_struct == multipliers_struct, (
        'multipliers should have same struct as params')
    return None

  def update_fn(updates, _, params=None):
    del params  # Unused.
    scaled_updates = jax.tree_multimap(lambda a, g: a * g, multipliers, updates)
    return scaled_updates, None

  return optax.GradientTransformation(init_fn, update_fn)


################ Pytrees / Misc ##################


def flatten(pytree, backend=np):
  """Take pytree of arrays, then flatten tree, flatten arrays, concatenate."""
  seq = tree.flatten(pytree)
  seq_flat = [backend.reshape(x, -1) for x in seq]
  return backend.concatenate(seq_flat)


def unflatten_like(a, pytree):
  """Take 1-D array produced by flatten() and unflatten like pytree."""
  seq = tree.flatten(pytree)
  seq_sizes = [np.reshape(x, -1).shape for x in seq]
  starts = [0] + list(np.cumsum(seq_sizes))
  a_seq_flat = [a[starts[i]:starts[i+1]] for i in range(len(starts)-1)]
  a_seq = [np.reshape(x1, x2.shape) for x1, x2 in zip(a_seq_flat, seq)]
  return tree.unflatten_as(pytree, a_seq)


def structure_like(tree1, tree2):
  # pylint: disable=g-doc-args, g-doc-return-or-yield
  """Makes tree1 have same structure as tree2."""
  flat_paths1 = tree.flatten_with_path(tree.map_structure(lambda x: 0, tree1))
  flat_paths2 = tree.flatten_with_path(tree.map_structure(lambda x: 0, tree2))
  assert list(sorted(flat_paths1)) == list(sorted(flat_paths2)), (
      'paths of tree1 and tree2 do not match')
  indices = [flat_paths1.index(path) for path in flat_paths2]
  flat_tree1 = tree.flatten(tree1)
  reordered_flat_tree1 = [flat_tree1[i] for i in indices]
  return tree.unflatten_as(tree2, reordered_flat_tree1)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup for pip package."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from setuptools import find_namespace_packages
from setuptools import setup


def _parse_requirements(requirements_txt_path):
  with open(requirements_txt_path) as fp:
    return fp.read().splitlines()


def test_suite():
  test_loader = unittest.TestLoader()
  all_tests = test_loader.discover('jax_verify/tests',
                                   pattern='*_test.py')
  return all_tests

setup(
    name='jax_verify',
    version='1.0',
    description='A library for neural network verification.',
    url='https://github.com/deepmind/jax_verify',
    author='DeepMind',
    author_email='jax_verify@google.com',
    # Contained modules and scripts.
    packages=find_namespace_packages(exclude=['*_test.py']),
    install_requires=_parse_requirements('requirements.txt'),
    requires_python='>=3.6',
    platforms=['any'],
    license='Apache 2.0',
    test_suite='setup.test_suite',
    include_package_data=True,
    zip_safe=False,
)



# coding=utf-8
# Copyright 2020 The jax_verify Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
r"""Run SDP verification for adversarial robustness specification.

Example launch commands which achieve good results:

CIFAR10 CNN-Mix:
python3 run_sdp_verify.py --model_name=models/cifar10_wongsmall_eps2_mix.pkl \
  --anneal_lengths=30000,30000,30000

MNIST CNN-Adv:
python3 run_sdp_verify.py --model_name=models/mnist_wongsmall_eps_10_adv.pkl \
  --epsilon=0.1 --dataset=mnist \
  --anneal_lengths=20000,20000,20000 --opt_name=adam --anneal_factor=0.03 \
  --n_iter_lanczos=300

MNIST Adv-MLP:
python3 run_sdp_verify.py --epsilon=0.1 --dataset=mnist \
  --model_name=models/raghunathan18_pgdnn.pkl --use_exact_eig_train=True \
  --use_exact_eig_eval=True --opt_name=adam --lam_coeff=0.1 --nu_coeff=0.03 \
  --custom_kappa_coeff=10000 --anneal_lengths=10000,4000,1000 \
  --kappa_zero_after=2000
"""


import functools
import os
import pickle

from absl import app
from absl import flags
import jax
import jax.numpy as jnp
import jax_verify
from jax_verify import sdp_verify
from jax_verify.src.sdp_verify import crown_boundprop
from jax_verify.src.sdp_verify import utils
import numpy as np

flags.DEFINE_integer('dataset_idx', 1, 'i^th example in dataset')
flags.DEFINE_integer('target_label_idx', 0, 'which class to target?')
flags.DEFINE_float('epsilon', 2. / 255, 'attack radius')
flags.DEFINE_string('dataset', 'cifar10', 'dataset, mnist or cifar')
flags.DEFINE_string('model_name', 'models/cifar10_wongsmall_eps2_mix.pkl',
                    'model name specifying Pickle file with network weights')
flags.DEFINE_boolean('inception_preprocess', False,
                     'Use inception_preprocessing i.e. [-1,1]-scaled inputs')
flags.DEFINE_boolean('use_crown', True,
                     'Use CROWN to compute activation bounds.')
flags.DEFINE_float('lam_coeff', 1.0, 'Coeff for dual variables')
flags.DEFINE_float('nu_coeff', 0.03, 'Coeff for dual variables')
flags.DEFINE_float('custom_kappa_coeff', -1,
                   'if >0, scale LR for top-left kappa')
flags.DEFINE_string('anneal_lengths', '15,5',
                    'comma-separated integers with # of steps per epoch')

# Flags passed directly to solver
flags.DEFINE_boolean('use_exact_eig_train', False,
                     'Use exact eigendecomposition for training')
flags.DEFINE_boolean('use_exact_eig_eval', False,
                     'Use exact eigendecomposition for evaluation')
flags.DEFINE_integer('n_iter_lanczos', 200, '# of Lanczos iters per step')
flags.DEFINE_float('eval_every', 1000, 'Iterations per log.')
flags.DEFINE_float('lr_init', 1e-3, 'initial learning rate')
flags.DEFINE_float('anneal_factor', 0.1, 'learning rate anneal factor')
flags.DEFINE_string('opt_name', 'rmsprop',
                    'Optix class: "adam" "sgd" or "rmsprop"')
flags.DEFINE_float('kappa_zero_after', 1e9, 'zero kappa_{1:n} after N steps')
flags.DEFINE_float('kappa_reg_weight', -1, '-1 disables kappa regularization')
FLAGS = flags.FLAGS


def _load_dataset(dataset):
  """Loads the 10000 MNIST (CIFAR) test set examples, saved as numpy arrays."""
  assert dataset in ('mnist', 'cifar10'), 'invalid dataset name'
  with jax_verify.open_file(os.path.join(dataset, 'x_test.npy'), 'rb') as f:
    xs = np.load(f)
  with jax_verify.open_file(os.path.join(dataset, 'y_test.npy'), 'rb') as f:
    ys = np.load(f)
  return xs, ys


def _load_weights(path):
  with jax_verify.open_file(path, 'rb') as f:
    data = pickle.load(f)
  return data


def get_verif_instance(params, x, label, target_label, epsilon,
                       input_bounds=(0., 1.)):
  """Creates verif instance."""
  if FLAGS.use_crown:
    bounds = crown_boundprop.boundprop(
        params, np.expand_dims(x, axis=0), epsilon, input_bounds=input_bounds)
  else:
    bounds = utils.boundprop(
        params, utils.init_bound(x, epsilon, input_bounds=input_bounds))
  verif_instance = utils.make_relu_robust_verif_instance(
      params, bounds, target_label=target_label, label=label,
      input_bounds=input_bounds)
  return verif_instance


def _opt_multiplier_fn(path, kappa_index, kappa_dim=None):
  """Set adaptive learning rates."""
  if FLAGS.custom_kappa_coeff > 0:
    kappa_lr_mul = FLAGS.custom_kappa_coeff
    if kappa_index in path:
      onehot = jax.nn.one_hot([0], kappa_dim)
      return jax.ops.index_update(onehot, (0, 0), kappa_lr_mul)
  if 'lam' in path:
    return FLAGS.lam_coeff
  if path == (kappa_index - 1, 'nu'):
    return FLAGS.nu_coeff
  return 1.0


def verify_cnn_single_dual(verif_instance):
  """Run verification for a CNN on a single MNIST/CIFAR problem."""
  verif_instance = utils.make_sdp_verif_instance(verif_instance)
  solver_params = dict(
      use_exact_eig_train=FLAGS.use_exact_eig_train,
      use_exact_eig_eval=FLAGS.use_exact_eig_eval,
      n_iter_lanczos=FLAGS.n_iter_lanczos,
      eval_every=FLAGS.eval_every,
      opt_name=FLAGS.opt_name,
      anneal_factor=FLAGS.anneal_factor,
      lr_init=FLAGS.lr_init,
      kappa_zero_after=FLAGS.kappa_zero_after,
      kappa_reg_weight=FLAGS.kappa_reg_weight,
  )
  # Set schedule
  steps_per_anneal = [int(x) for x in FLAGS.anneal_lengths.split(',')]
  num_steps = sum(steps_per_anneal)
  solver_params['steps_per_anneal'] = steps_per_anneal[:-1] + [int(1e9)]

  # Set learning rate multipliers
  kappa_shape = verif_instance.dual_shapes[-1]
  kappa_index = len(verif_instance.dual_shapes) - 1
  assert len(kappa_shape) == 2 and kappa_shape[0] == 1
  opt_multiplier_fn = functools.partial(
      _opt_multiplier_fn, kappa_index=kappa_index, kappa_dim=kappa_shape[1])

  # Call solver
  obj_value, info = sdp_verify.solve_sdp_dual(
      verif_instance, num_steps=num_steps, verbose=True,
      opt_multiplier_fn=opt_multiplier_fn, **solver_params)
  info['final_dual_vars'] = jax.tree_map(np.array, info['final_dual_vars'])
  return float(obj_value), info


class PickleWriter:

  def write(self, d):
    with open('/tmp/run_sdp_verify_results.pkl', 'wb') as f:
      pickle.dump(d, f)


def main(unused_argv):
  run_verification(PickleWriter())


def run_verification(writer):
  """Run verification."""
  xs, ys = _load_dataset(FLAGS.dataset)
  dataset_idx = FLAGS.dataset_idx
  if FLAGS.dataset == 'cifar10':
    x = utils.preprocess_cifar(xs[dataset_idx])
    epsilon, input_bounds = utils.preprocessed_cifar_eps_and_input_bounds(
        shape=x.shape, epsilon=FLAGS.epsilon,
        inception_preprocess=FLAGS.inception_preprocess)
  else:
    x = xs[dataset_idx]
    epsilon = FLAGS.epsilon
    input_bounds = (0., 1.)
  true_label = ys[dataset_idx]
  target_label = FLAGS.target_label_idx
  params = _load_weights(FLAGS.model_name)
  if isinstance(params[0], dict):
    params[0]['input_shape'] = x.shape[0]

  verif_instance = get_verif_instance(
      params, x, label=true_label, target_label=target_label,
      epsilon=epsilon, input_bounds=input_bounds)

  # Report initial bound from interval bounds.
  ibp_bound = utils.ibp_bound_elided(verif_instance)
  print('IBP bound:', ibp_bound)
  if true_label == target_label:
    return

  # Run dual SDP verification.
  verified_ub, info = verify_cnn_single_dual(verif_instance)

  # Run FGSM eval.
  model_fn = lambda x: utils.predict_cnn(params, jnp.expand_dims(x, axis=0))
  x_adv = utils.fgsm_single(
      model_fn, x, true_label, target_label, input_bounds=input_bounds,
      epsilon=epsilon, num_steps=100, step_size=0.001)
  adv_objective = float(
      utils.adv_objective(model_fn, x_adv, true_label, target_label))
  print('adv_objective :', adv_objective)

  output_dict = {
      'dataset_idx': dataset_idx,
      'true_label': true_label,
      'target_label': target_label,
      'epsilon': FLAGS.epsilon,
      'verified_ub': verified_ub,
      'adv_lb': adv_objective,
      'adv_success': adv_objective > 0.0,
      'ibp_bound': ibp_bound,
  }
  output_dict.update(info)
  jax_to_np = lambda x: np.array(x) if isinstance(x, jnp.DeviceArray) else x
  output_dict = jax.tree_map(jax_to_np, output_dict)
  writer.write(output_dict)

if __name__ == '__main__':
  app.run(main)
